from __future__ import annotations

import io
import os
import weakref
from contextlib import suppress
from typing import Any

from typing_extensions import Self

from .. import optional_features
from ..element import Element

with suppress(ImportError):
    if os.environ.get('MATPLOTLIB', 'true').lower() == 'true':
        import matplotlib.figure
        import matplotlib.pyplot as plt
        optional_features.register('matplotlib')

        class MatplotlibFigure(matplotlib.figure.Figure):

            def __init__(self, element: Matplotlib, *args: Any, **kwargs: Any) -> None:
                super().__init__(*args, **kwargs)
                self._element = weakref.ref(element)

            @property
            def element(self) -> Matplotlib:
                """The element this matplotlib figure belongs to."""
                element = self._element()
                if element is None:
                    raise RuntimeError('The element this matplotlib figure belongs to has been deleted.')
                return element

            def __enter__(self) -> Self:
                return self

            def __exit__(self, *_) -> None:
                self.element.update()


class Pyplot(Element, default_classes='nicegui-pyplot'):

    def __init__(self, *, close: bool = True, **kwargs: Any) -> None:
        """Pyplot Context

        Create a context to configure a `Matplotlib <https://matplotlib.org/>`_ plot.

        :param close: whether the figure should be closed after exiting the context; set to `False` if you want to update it later (default: `True`)
        :param kwargs: arguments like `figsize` which should be passed to `pyplot.figure <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.figure.html>`_
        """
        if not optional_features.has('matplotlib'):
            raise ImportError('Matplotlib is not installed. Please run "pip install matplotlib".')

        super().__init__('div')
        self.close = close
        self.fig = plt.figure(**kwargs)  # pylint: disable=possibly-used-before-assignment
        self._convert_to_html()

    def _convert_to_html(self) -> None:
        with io.StringIO() as output:
            self.fig.savefig(output, format='svg')
            self._props['innerHTML'] = output.getvalue()

    def __enter__(self) -> Self:
        plt.figure(self.fig)
        return self

    def __exit__(self, *_) -> None:
        self._convert_to_html()
        if self.close:
            plt.close(self.fig)
        self.update()

    def _handle_delete(self) -> None:
        plt.close(self.fig)
        super()._handle_delete()


class Matplotlib(Element, default_classes='nicegui-matplotlib'):

    def __init__(self, **kwargs: Any) -> None:
        """Matplotlib

        Create a `Matplotlib <https://matplotlib.org/>`_ element rendering a Matplotlib figure.
        The figure is automatically updated when leaving the figure context.

        :param kwargs: arguments like `figsize` which should be passed to `matplotlib.figure.Figure <https://matplotlib.org/stable/api/figure_api.html#matplotlib.figure.Figure>`_
        """
        if not optional_features.has('matplotlib'):
            raise ImportError('Matplotlib is not installed. Please run "pip install matplotlib".')

        super().__init__('div')
        self.figure = MatplotlibFigure(self, **kwargs)  # pylint: disable=possibly-used-before-assignment
        self._convert_to_html()

    def _convert_to_html(self) -> None:
        with io.StringIO() as output:
            self.figure.savefig(output, format='svg')
            self._props['innerHTML'] = output.getvalue()

    def update(self) -> None:
        with self._props.suspend_updates():
            self._convert_to_html()
        super().update()



import weakref

from typing_extensions import Self

from ..classes import Classes
from ..context import context
from ..element import Element
from ..props import Props
from ..style import Style


class QueryElement(Element, component='query.js'):

    def __init__(self, selector: str) -> None:
        super().__init__()
        self._props['selector'] = selector
        self._props['classes'] = []
        self._props['style'] = {}
        self._props['props'] = {}


class Query:

    def __init__(self, selector: str) -> None:
        """Query Selector

        To manipulate elements like the document body, you can use the `ui.query` function.
        With the query result you can add classes, styles, and attributes like with every other UI element.
        This can be useful for example to change the background color of the page (e.g. `ui.query('body').classes('bg-green')`).

        :param selector: the CSS selector (e.g. "body", "#my-id", ".my-class", "div > p")
        """
        for element in context.client.elements.values():
            if isinstance(element, QueryElement) and element.props['selector'] == selector:
                self._element = weakref.ref(element)
                break
        else:
            self._element = weakref.ref(QueryElement(selector))

    @property
    def element(self) -> QueryElement:
        """The element this query belongs to."""
        element = self._element()
        if element is None:
            raise RuntimeError('The element this query belongs to has been deleted.')
        return element

    def classes(self,
                add: str | None = None, *,
                remove: str | None = None,
                toggle: str | None = None,
                replace: str | None = None,
                ) -> Self:
        """Apply, remove, toggle, or replace HTML classes.

        This allows modifying the look of the element or its layout using `Tailwind <https://tailwindcss.com/>`_ or `Quasar <https://quasar.dev/>`_ classes.

        Removing or replacing classes can be helpful if predefined classes are not desired.

        :param add: whitespace-delimited string of classes
        :param remove: whitespace-delimited string of classes to remove from the element
        :param toggle: whitespace-delimited string of classes to toggle (*added in version 2.7.0*)
        :param replace: whitespace-delimited string of classes to use instead of existing ones
        """
        element = self.element
        classes = Classes.update_list(element.props['classes'], add, remove, toggle, replace)
        new_classes = [c for c in classes if c not in element.props['classes']]
        old_classes = [c for c in element.props['classes'] if c not in classes]
        if new_classes:
            element.run_method('add_classes', new_classes)
        if old_classes:
            element.run_method('remove_classes', old_classes)
        element.props['classes'] = classes
        return self

    def style(self, add: str | None = None, *, remove: str | None = None, replace: str | None = None) \
            -> Self:
        """Apply, remove, or replace CSS definitions.

        Removing or replacing styles can be helpful if the predefined style is not desired.

        :param add: semicolon-separated list of styles to add to the element
        :param remove: semicolon-separated list of styles to remove from the element
        :param replace: semicolon-separated list of styles to use instead of existing ones
        """
        element = self.element
        old_style = Style.parse(remove)
        for key in old_style:
            element.props['style'].pop(key, None)
        if old_style:
            element.run_method('remove_style', list(old_style))
        element.props['style'].update(Style.parse(add))
        element.props['style'].update(Style.parse(replace))
        if element.props['style']:
            element.run_method('add_style', element.props['style'])
        return self

    def props(self, add: str | None = None, *, remove: str | None = None) -> Self:
        """Add or remove props.

        This allows modifying the look of the element or its layout using `Quasar <https://quasar.dev/>`_ props.
        Since props are simply applied as HTML attributes, they can be used with any HTML element.

        Boolean properties are assumed ``True`` if no value is specified.

        :param add: whitespace-delimited list of either boolean values or key=value pair to add
        :param remove: whitespace-delimited list of property keys to remove
        """
        element = self.element
        old_props = Props.parse(remove)
        for key in old_props:
            element.props['props'].pop(key, None)
        if old_props:
            element.run_method('remove_props', list(old_props))
        new_props = Props.parse(add)
        element.props['props'].update(new_props)
        if element.props['props']:
            element.run_method('add_props', element.props['props'])
        return self



import os

import ifaddr

from . import core, run


def _get_all_ips() -> list[str]:
    ips: list[str] = []
    for adapter in ifaddr.get_adapters():
        ips.extend(str(i.ip) for i in adapter.ips if i.is_IPv4)
    return ips


async def collect_urls() -> None:
    """Print a welcome message with URLs to access the NiceGUI app."""
    host = os.environ.get('NICEGUI_HOST')
    port = os.environ.get('NICEGUI_PORT')
    protocol = os.environ.get('NICEGUI_PROTOCOL')
    if not host or not port or not protocol:
        return
    ips = set((await run.io_bound(_get_all_ips)) if host == '0.0.0.0' else [])
    ips.discard('127.0.0.1')
    sorted_ips = ['localhost' if host == '0.0.0.0' else host, *sorted(ips)]
    urls = [(f'{protocol}://{ip}:{port}' if port != '80' else f'{protocol}://{ip}') for ip in sorted_ips]
    core.app.urls.update(urls)
    if len(urls) >= 2:
        urls[-1] = 'and ' + urls[-1]
    if core.app.config.show_welcome_message:
        print(f'NiceGUI ready to go on {", ".join(urls)}', flush=True)



from __future__ import annotations

from contextlib import suppress

from ... import optional_features
from ...element import Element

with suppress(ImportError):
    import plotly.graph_objects as go
    optional_features.register('plotly')


class Plotly(Element, component='plotly.js', esm={'nicegui-plotly': 'dist'}):

    def __init__(self, figure: dict | go.Figure) -> None:
        """Plotly Element

        Renders a Plotly chart.
        There are two ways to pass a Plotly figure for rendering, see parameter `figure`:

        * Pass a `go.Figure` object, see https://plotly.com/python/

        * Pass a Python `dict` object with keys `data`, `layout`, `config` (optional), see https://plotly.com/javascript/

        For best performance, use the declarative `dict` approach for creating a Plotly chart.

        :param figure: Plotly figure to be rendered. Can be either a `go.Figure` instance, or
                       a `dict` object with keys `data`, `layout`, `config` (optional).
        """
        super().__init__()

        self.figure = figure
        self.update()
        self._classes.append('js-plotly-plot')
        self._update_method = 'update'

    def update_figure(self, figure: dict | go.Figure):
        """Overrides figure instance of this Plotly chart and updates chart on client side."""
        self.figure = figure
        self.update()

    def update(self) -> None:
        with self._props.suspend_updates():
            self._props['options'] = self._get_figure_json()
        super().update()

    def _get_figure_json(self) -> dict:
        if optional_features.has('plotly') and isinstance(self.figure, go.Figure):
            # convert go.Figure to dict object which is directly JSON serializable
            # orjson supports NumPy array serialization
            return self.figure.to_plotly_json()

        if isinstance(self.figure, dict):
            # already a dict object with keys: data, layout, config (optional)
            return self.figure

        raise ValueError(f'Plotly figure is of unknown type "{self.figure.__class__.__name__}".')



from ..element import Element


class ButtonGroup(Element):

    def __init__(self) -> None:
        """Button Group

        This element is based on Quasar's `QBtnGroup <https://quasar.dev/vue-components/button-group>`_ component.
        You must use the same design props on both the parent button group and the children buttons.
        """
        super().__init__(tag='q-btn-group')



from __future__ import annotations

import abc
import time
from collections.abc import Callable, Collection, Iterable
from copy import deepcopy
from typing import Any, SupportsIndex

from typing_extensions import Self

from . import events


class ObservableCollection(abc.ABC):  # noqa: B024

    def __init__(self, *,
                 factory: Callable,
                 data: Collection | None,
                 on_change: Callable | None,
                 _parent: ObservableCollection | None,
                 ) -> None:
        super().__init__(factory() if data is None else data)  # type: ignore
        self._parent = _parent
        self.last_modified = time.time()
        self._change_handlers: list[Callable] = [on_change] if on_change else []

    @property
    def change_handlers(self) -> list[Callable]:
        """Return a list of all change handlers registered on this collection and its parents."""
        change_handlers = self._change_handlers[:]
        if self._parent is not None:
            change_handlers.extend(self._parent.change_handlers)
        return change_handlers

    def _handle_change(self) -> None:
        self.last_modified = time.time()
        for handler in self.change_handlers:
            events.handle_event(handler, events.ObservableChangeEventArguments(sender=self))

    def on_change(self, handler: Callable) -> None:
        """Register a handler to be called when the collection changes."""
        if handler != self._handle_change:  # pylint: disable=comparison-with-callable
            self._change_handlers.append(handler)

    def _observe(self, data: Any) -> Any:
        if isinstance(data, ObservableCollection):
            data.on_change(self._handle_change)
            return data
        if isinstance(data, dict):
            return ObservableDict(data, _parent=self)
        if isinstance(data, list):
            return ObservableList(data, _parent=self)
        if isinstance(data, set):
            return ObservableSet(data, _parent=self)
        return data

    def __copy__(self) -> Self:
        if isinstance(self, dict):
            return ObservableDict(self, _parent=self._parent)
        if isinstance(self, list):
            return ObservableList(self, _parent=self._parent)
        if isinstance(self, set):
            return ObservableSet(self, _parent=self._parent)
        raise NotImplementedError(f'ObservableCollection.__copy__ not implemented for {type(self)}')

    def __deepcopy__(self, memo: dict) -> Self:
        if isinstance(self, dict):
            return ObservableDict({key: deepcopy(self[key]) for key in self}, _parent=self._parent)
        if isinstance(self, list):
            return ObservableList([deepcopy(item) for item in self], _parent=self._parent)
        if isinstance(self, set):
            return ObservableSet({deepcopy(item) for item in self}, _parent=self._parent)
        raise NotImplementedError(f'ObservableCollection.__deepcopy__ not implemented for {type(self)}')


class ObservableDict(ObservableCollection, dict):

    def __init__(self,
                 data: dict | None = None,
                 *,
                 on_change: Callable | None = None,
                 _parent: ObservableCollection | None = None,
                 ) -> None:
        super().__init__(factory=dict, data=data, on_change=on_change, _parent=_parent)
        for key, value in self.items():
            super().__setitem__(key, self._observe(value))

    def pop(self, k: Any, d: Any = None) -> Any:
        item = super().pop(k, d)
        self._handle_change()
        return item

    def popitem(self) -> Any:
        item = super().popitem()
        self._handle_change()
        return item

    def update(self, *args: Any, **kwargs: Any) -> None:
        super().update(self._observe(dict(*args, **kwargs)))
        self._handle_change()

    def clear(self) -> None:
        super().clear()
        self._handle_change()

    def setdefault(self, __key: Any, __default: Any = None) -> Any:
        item = super().setdefault(__key, self._observe(__default))
        self._handle_change()
        return item

    def __setitem__(self, __key: Any, __value: Any) -> None:
        super().__setitem__(__key, self._observe(__value))
        self._handle_change()

    def __delitem__(self, __key: Any) -> None:
        super().__delitem__(__key)
        self._handle_change()

    def __or__(self, other: Any) -> Any:
        return super().__or__(other)

    def __ior__(self, other: Any) -> Any:
        other_dict = self._observe(dict(other))
        super().__ior__(other_dict)
        self._handle_change()
        return self


class ObservableList(ObservableCollection, list):

    def __init__(self,
                 data: list | None = None,
                 *,
                 on_change: Callable | None = None,
                 _parent: ObservableCollection | None = None,
                 ) -> None:
        super().__init__(factory=list, data=data, on_change=on_change, _parent=_parent)
        for i, item in enumerate(self):
            super().__setitem__(i, self._observe(item))

    def append(self, item: Any) -> None:
        super().append(self._observe(item))
        self._handle_change()

    def extend(self, iterable: Iterable) -> None:
        super().extend(self._observe(list(iterable)))
        self._handle_change()

    def insert(self, index: SupportsIndex, obj: Any) -> None:
        super().insert(index, self._observe(obj))
        self._handle_change()

    def remove(self, value: Any) -> None:
        super().remove(value)
        self._handle_change()

    def pop(self, index: SupportsIndex = -1) -> Any:
        item = super().pop(index)
        self._handle_change()
        return item

    def clear(self) -> None:
        super().clear()
        self._handle_change()

    def sort(self, **kwargs: Any) -> None:
        super().sort(**kwargs)
        self._handle_change()

    def reverse(self) -> None:
        super().reverse()
        self._handle_change()

    def __delitem__(self, key: SupportsIndex | slice) -> None:
        super().__delitem__(key)
        self._handle_change()

    def __setitem__(self, key: SupportsIndex | slice, value: Any) -> None:
        super().__setitem__(key, self._observe(value))
        self._handle_change()

    def __add__(self, other: Any) -> Any:
        return super().__add__(other)

    def __iadd__(self, other: Any) -> Any:
        super().__iadd__(self._observe(other))
        self._handle_change()
        return self


class ObservableSet(ObservableCollection, set):

    def __init__(self,
                 data: set | None = None,
                 *,
                 on_change: Callable | None = None,
                 _parent: ObservableCollection | None = None,
                 ) -> None:
        super().__init__(factory=set, data=data, on_change=on_change, _parent=_parent)
        for item in self:
            super().add(self._observe(item))

    def add(self, item: Any) -> None:
        super().add(self._observe(item))
        self._handle_change()

    def remove(self, item: Any) -> None:
        super().remove(item)
        self._handle_change()

    def discard(self, item: Any) -> None:
        super().discard(item)
        self._handle_change()

    def pop(self) -> Any:
        item = super().pop()
        self._handle_change()
        return item

    def clear(self) -> None:
        super().clear()
        self._handle_change()

    def update(self, *s: Iterable[Any]) -> None:
        super().update(self._observe(set(*s)))
        self._handle_change()

    def intersection_update(self, *s: Iterable[Any]) -> None:
        super().intersection_update(*s)
        self._handle_change()

    def difference_update(self, *s: Iterable[Any]) -> None:
        super().difference_update(*s)
        self._handle_change()

    def symmetric_difference_update(self, *s: Iterable[Any]) -> None:
        super().symmetric_difference_update(*s)
        self._handle_change()

    def __or__(self, other: Any) -> Any:
        return super().__or__(other)

    def __ior__(self, other: Any) -> Any:
        super().__ior__(self._observe(other))
        self._handle_change()
        return self

    def __and__(self, other: Any) -> set:
        return super().__and__(other)

    def __iand__(self, other: Any) -> Any:
        super().__iand__(self._observe(other))
        self._handle_change()
        return self

    def __sub__(self, other: Any) -> set:
        return super().__sub__(other)

    def __isub__(self, other: Any) -> Any:
        super().__isub__(self._observe(other))
        self._handle_change()
        return self

    def __xor__(self, other: Any) -> set:
        return super().__xor__(other)

    def __ixor__(self, other: Any) -> Any:
        super().__ixor__(self._observe(other))
        self._handle_change()
        return self



from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response

from . import core
from .version import __version__


class RedirectWithPrefixMiddleware(BaseHTTPMiddleware):

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        prefix = request.headers.get('X-Forwarded-Prefix', '')
        response = await call_next(request)
        if 'Location' in response.headers and response.headers['Location'].startswith('/'):
            new_location = prefix + response.headers['Location']
            response.headers['Location'] = new_location
        return response


class SetCacheControlMiddleware(BaseHTTPMiddleware):

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        response = await call_next(request)
        if request.url.path.startswith(f'/_nicegui/{__version__}/') \
                and not request.url.path.startswith(f'/_nicegui/{__version__}/dynamic_resources/'):
            response.headers['Cache-Control'] = core.app.config.cache_control_directives
        return response



from __future__ import annotations

from typing import Any

from ..context import context
from ..defaults import DEFAULT_PROP, resolve_defaults
from ..events import Handler, ValueChangeEventArguments
from .mixins.disableable_element import DisableableElement
from .mixins.icon_element import IconElement
from .mixins.label_element import LabelElement
from .mixins.value_element import ValueElement


class Tabs(ValueElement):

    def __init__(self, *,
                 value: Tab | TabPanel | None = None,
                 on_change: Handler[ValueChangeEventArguments] | None = None,
                 ) -> None:
        """Tabs

        This element represents `Quasar's QTabs <https://quasar.dev/vue-components/tabs#qtabs-api>`_ component.
        It contains individual tabs.

        :param value: `ui.tab`, `ui.tab_panel`, or name of the tab to be initially selected
        :param on_change: callback to be executed when the selected tab changes (*since version 3.6.0*: event ``value`` is the tab name)
        """
        super().__init__(tag='q-tabs', value=value, on_value_change=on_change)

    def _value_to_model_value(self, value: Any) -> Any:
        return value.props['name'] if isinstance(value, (Tab, TabPanel)) else value

    def _value_to_event_value(self, value: Any) -> Any:
        return self._value_to_model_value(value)


class Tab(LabelElement, IconElement, DisableableElement):

    def __init__(self, name: str, label: str | None = None, icon: str | None = None) -> None:
        """Tab

        This element represents `Quasar's QTab <https://quasar.dev/vue-components/tabs#qtab-api>`_ component.
        It is a child of a `ui.tabs` element.

        :param name: name of the tab (will be the value of the `ui.tabs` element)
        :param label: label of the tab (default: `None`, meaning the same as `name`)
        :param icon: icon of the tab (default: `None`)
        """
        if label is None:
            label = name
        super().__init__(tag='q-tab', label=label, icon=icon)
        self._props['name'] = name
        self.tabs = context.slot.parent


class TabPanels(ValueElement):

    @resolve_defaults
    def __init__(self,
                 tabs: Tabs | None = None, *,
                 value: Tab | TabPanel | str | None = None,
                 on_change: Handler[ValueChangeEventArguments] | None = None,
                 animated: bool = DEFAULT_PROP | True,
                 keep_alive: bool = DEFAULT_PROP | True,
                 ) -> None:
        """Tab Panels

        This element represents `Quasar's QTabPanels <https://quasar.dev/vue-components/tab-panels#qtabpanels-api>`_ component.
        It contains individual tab panels.

        To avoid issues with dynamic elements when switching tabs,
        this element uses Vue's `keep-alive <https://vuejs.org/guide/built-ins/keep-alive.html>`_ component.
        If client-side performance is an issue, you can disable this feature.

        :param tabs: an optional `ui.tabs` element that controls this element
        :param value: `ui.tab`, `ui.tab_panel`, or name of the tab panel to be initially visible
        :param on_change: callback to be executed when the visible tab panel changes (*since version 3.6.0*: event ``value`` is the tab name)
        :param animated: whether the tab panels should be animated (default: `True`)
        :param keep_alive: whether to use Vue's keep-alive component on the content (default: `True`)
        """
        super().__init__(tag='q-tab-panels', value=value, on_value_change=on_change)
        if tabs is not None:
            tabs.bind_value(self, 'value')
        self._props.set_bool('animated', animated)
        self._props.set_bool('keep-alive', keep_alive)

    def _value_to_model_value(self, value: Any) -> Any:
        return value.props['name'] if isinstance(value, (Tab, TabPanel)) else value

    def _value_to_event_value(self, value: Any) -> Any:
        return self._value_to_model_value(value)


class TabPanel(DisableableElement, default_classes='nicegui-tab-panel'):

    def __init__(self, name: Tab | str) -> None:
        """Tab Panel

        This element represents `Quasar's QTabPanel <https://quasar.dev/vue-components/tab-panels#qtabpanel-api>`_ component.
        It is a child of a `TabPanels` element.

        :param name: `ui.tab` or the name of a tab element
        """
        super().__init__(tag='q-tab-panel')
        self._props['name'] = name.props['name'] if isinstance(name, Tab) else name



import math

from .scene_object3d import Object3D


class Group(Object3D):

    def __init__(self) -> None:
        """Group

        This element is based on Three.js' `Group <https://threejs.org/docs/index.html#api/en/objects/Group>`_ object.
        It is used to group objects together.
        """
        super().__init__('group')


class Box(Object3D):

    def __init__(self,
                 width: float = 1.0,
                 height: float = 1.0,
                 depth: float = 1.0,
                 wireframe: bool = False,
                 ) -> None:
        """Box

        This element is based on Three.js' `BoxGeometry <https://threejs.org/docs/index.html#api/en/geometries/BoxGeometry>`_ object.
        It is used to create a box-shaped mesh.

        :param width: width of the box (default: 1.0)
        :param height: height of the box (default: 1.0)
        :param depth: depth of the box (default: 1.0)
        :param wireframe: whether to display the box as a wireframe (default: `False`)
        """
        super().__init__('box', width, height, depth, wireframe)


class Sphere(Object3D):

    def __init__(self,
                 radius: float = 1.0,
                 width_segments: int = 32,
                 height_segments: int = 16,
                 wireframe: bool = False,
                 ) -> None:
        """Sphere

        This element is based on Three.js' `SphereGeometry <https://threejs.org/docs/index.html#api/en/geometries/SphereGeometry>`_ object.
        It is used to create a sphere-shaped mesh.

        :param radius: radius of the sphere (default: 1.0)
        :param width_segments: number of horizontal segments (default: 32)
        :param height_segments: number of vertical segments (default: 16)
        :param wireframe: whether to display the sphere as a wireframe (default: `False`)
        """
        super().__init__('sphere', radius, width_segments, height_segments, wireframe)


class Cylinder(Object3D):

    def __init__(self,
                 top_radius: float = 1.0,
                 bottom_radius: float = 1.0,
                 height: float = 1.0,
                 radial_segments: int = 8,
                 height_segments: int = 1,
                 wireframe: bool = False,
                 ) -> None:
        """Cylinder

        This element is based on Three.js' `CylinderGeometry <https://threejs.org/docs/index.html#api/en/geometries/CylinderGeometry>`_ object.
        It is used to create a cylinder-shaped mesh.

        :param top_radius: radius of the top (default: 1.0)
        :param bottom_radius: radius of the bottom (default: 1.0)
        :param height: height of the cylinder (default: 1.0)
        :param radial_segments: number of horizontal segments (default: 8)
        :param height_segments: number of vertical segments (default: 1)
        :param wireframe: whether to display the cylinder as a wireframe (default: `False`)
        """
        super().__init__('cylinder', top_radius, bottom_radius, height, radial_segments, height_segments, wireframe)


class Ring(Object3D):

    def __init__(self,
                 inner_radius: float = 0.5,
                 outer_radius: float = 1.0,
                 theta_segments: int = 8,
                 phi_segments: int = 1,
                 theta_start: float = 0,
                 theta_length: float = 2 * math.pi,
                 wireframe: bool = False,
                 ) -> None:
        """Ring

        This element is based on Three.js' `RingGeometry <https://threejs.org/docs/index.html#api/en/geometries/RingGeometry>`_ object.
        It is used to create a ring-shaped mesh.

        :param inner_radius: inner radius of the ring (default: 0.5)
        :param outer_radius: outer radius of the ring (default: 1.0)
        :param theta_segments: number of horizontal segments (default: 8, higher means rounder)
        :param phi_segments: number of vertical segments (default: 1)
        :param theta_start: start angle in radians (default: 0)
        :param theta_length: central angle in radians (default: 2π)
        :param wireframe: whether to display the ring as a wireframe (default: `False`)
        """
        super().__init__('ring',
                         inner_radius, outer_radius, theta_segments, phi_segments, theta_start, theta_length, wireframe)


class QuadraticBezierTube(Object3D):

    def __init__(self,
                 start: list[float],
                 mid: list[float],
                 end: list[float],
                 tubular_segments: int = 64,
                 radius: float = 1.0,
                 radial_segments: int = 8,
                 closed: bool = False,
                 wireframe: bool = False,
                 ) -> None:
        """Quadratic Bezier Tube

        This element is based on Three.js' `QuadraticBezierCurve3 <https://threejs.org/docs/index.html#api/en/extras/curves/QuadraticBezierCurve3>`_ object.
        It is used to create a tube-shaped mesh.

        :param start: start point of the curve
        :param mid: middle point of the curve
        :param end: end point of the curve
        :param tubular_segments: number of tubular segments (default: 64)
        :param radius: radius of the tube (default: 1.0)
        :param radial_segments: number of radial segments (default: 8)
        :param closed: whether the tube should be closed (default: `False`)
        :param wireframe: whether to display the tube as a wireframe (default: `False`)
        """
        super().__init__('quadratic_bezier_tube',
                         start, mid, end, tubular_segments, radius, radial_segments, closed, wireframe)


class Extrusion(Object3D):

    def __init__(self,
                 outline: list[list[float]],
                 height: float,
                 wireframe: bool = False,
                 ) -> None:
        """Extrusion

        This element is based on Three.js' `ExtrudeGeometry <https://threejs.org/docs/index.html#api/en/geometries/ExtrudeGeometry>`_ object.
        It is used to create a 3D shape by extruding a 2D shape to a given height.

        :param outline: list of points defining the outline of the 2D shape
        :param height: height of the extrusion
        :param wireframe: whether to display the extrusion as a wireframe (default: `False`)
        """
        super().__init__('extrusion', outline, height, wireframe)


class Stl(Object3D):

    def __init__(self,
                 url: str,
                 wireframe: bool = False,
                 ) -> None:
        """STL

        This element is used to create a mesh from an STL file.

        :param url: URL of the STL file
        :param wireframe: whether to display the STL as a wireframe (default: `False`)
        """
        super().__init__('stl', url, wireframe)


class Gltf(Object3D):

    def __init__(self,
                 url: str,
                 ) -> None:
        """GLTF

        This element is used to create a mesh from a glTF file.

        :param url: URL of the glTF file
        """
        super().__init__('gltf', url)


class Line(Object3D):

    def __init__(self,
                 start: list[float],
                 end: list[float],
                 ) -> None:
        """Line

        This element is based on Three.js' `Line <https://threejs.org/docs/index.html#api/en/objects/Line>`_ object.
        It is used to create a line segment.

        :param start: start point of the line
        :param end: end point of the line
        """
        super().__init__('line', start, end)


class Curve(Object3D):

    def __init__(self,
                 start: list[float],
                 control1: list[float],
                 control2: list[float],
                 end: list[float],
                 num_points: int = 20,
                 ) -> None:
        """Curve

        This element is based on Three.js' `CubicBezierCurve3 <https://threejs.org/docs/index.html#api/en/extras/curves/CubicBezierCurve3>`_ object.

        :param start: start point of the curve
        :param control1: first control point of the curve
        :param control2: second control point of the curve
        :param end: end point of the curve
        :param num_points: number of points to use for the curve (default: 20)
        """
        super().__init__('curve', start, control1, control2, end, num_points)


class Text(Object3D):

    def __init__(self,
                 text: str,
                 style: str = '',
                 ) -> None:
        """Text

        This element is used to add 2D text to the scene.
        It can be moved like any other object, but always faces the camera.

        :param text: text to display
        :param style: CSS style (default: '')
        """
        super().__init__('text', text, style)


class Text3d(Object3D):

    def __init__(self,
                 text: str,
                 style: str = '',
                 ) -> None:
        """3D Text

        This element is used to add a 3D text mesh to the scene.
        It can be moved and rotated like any other object.

        :param text: text to display
        :param style: CSS style (default: '')
        """
        super().__init__('text3d', text, style)


class Texture(Object3D):

    def __init__(self,
                 url: str,
                 coordinates: list[list[list[float] | None]],
                 ) -> None:
        """Texture

        This element is used to add a texture to a mesh.

        :param url: URL of the texture image
        :param coordinates: texture coordinates
        """
        super().__init__('texture', url, coordinates)

    def set_url(self, url: str) -> None:
        """Change the URL of the texture image."""
        self.args[0] = url
        self.scene.run_method('set_texture_url', self.id, url)

    def set_coordinates(self, coordinates: list[list[list[float] | None]]) -> None:
        """Change the texture coordinates."""
        self.args[1] = coordinates
        self.scene.run_method('set_texture_coordinates', self.id, coordinates)


class SpotLight(Object3D):

    def __init__(self,
                 color: str = '#ffffff',
                 intensity: float = 1.0,
                 distance: float = 0.0,
                 angle: float = math.pi / 3,
                 penumbra: float = 0.0,
                 decay: float = 1.0,
                 ) -> None:
        """Spot Light

        This element is based on Three.js' `SpotLight <https://threejs.org/docs/index.html#api/en/lights/SpotLight>`_ object.
        It is used to add a spot light to the scene.

        :param color: CSS color string (default: '#ffffff')
        :param intensity: light intensity (default: 1.0)
        :param distance: maximum distance of light (default: 0.0)
        :param angle: maximum angle of light (default: π/2)
        :param penumbra: penumbra (default: 0.0)
        :param decay: decay (default: 2.0)
        """
        super().__init__('spot_light', color, intensity, distance, angle, penumbra, decay)


class PointCloud(Object3D):

    def __init__(self,
                 points: list[list[float]],
                 colors: list[list[float]] | None = None,
                 point_size: float = 1.0,
                 ) -> None:
        """Point Cloud

        This element is based on Three.js' `Points <https://threejs.org/docs/index.html#api/en/objects/Points>`_ object.

        :param points: list of points
        :param colors: optional list of colors (one per point)
        :param point_size: size of the points (default: 1.0)
        """
        super().__init__('point_cloud', points, colors, point_size)
        if colors is not None:
            self.material(color=None)

    def set_points(self, points: list[list[float]], colors: list[list[float]] | None = None) -> None:
        """Change the points and colors of the point cloud."""
        self.args[0] = points
        self.args[1] = colors
        self.scene.run_method('set_points', self.id, points, colors)
        if colors is not None:
            self.material(color=None)


class AxesHelper(Object3D):

    def __init__(self,
                 length: float = 1.0,
                 ) -> None:
        """Axes Helper

        This element is based on Three.js' `AxesHelper <https://threejs.org/docs/#api/en/helpers/AxesHelper>`_ object.
        It is used to visualize the XYZ axes:
        The X axis is red.
        The Y axis is green.
        The Z axis is blue.

        :param length: length of the the axes (default: 1.0)
        """
        super().__init__('axes_helper', length)



from collections.abc import Callable
from pathlib import Path
from typing import Any, cast

from typing_extensions import Self

from ... import core
from ...binding import BindableProperty, bind, bind_from, bind_to
from ...element import Element
from ...helpers import is_file


class SourceElement(Element):
    source = BindableProperty(
        on_change=lambda sender, source: cast(Self, sender)._handle_source_change(source))  # pylint: disable=protected-access

    SOURCE_IS_MEDIA_FILE: bool = False

    def __init__(self, *, source: Any, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.auto_route: str | None = None
        self._temp_path: Path | None = None
        self.source = source
        self._set_props(source)

    def bind_source_to(self,
                       target_object: Any,
                       target_name: str = 'source',
                       forward: Callable[[Any], Any] | None = None, *,
                       strict: bool | None = None,
                       ) -> Self:
        """Bind the source of this element to the target object's target_name property.

        The binding works one way only, from this element to the target.
        The update happens immediately and whenever a value changes.

        :param target_object: The object to bind to.
        :param target_name: The name of the property to bind to.
        :param forward: A function to apply to the value before applying it to the target (default: identity).
        :param strict: Whether to check (and raise) if the target object has the specified property (default: None,
            performs a check if the object is not a dictionary, *added in version 3.0.0*).
        """
        bind_to(self, 'source', target_object, target_name, forward, self_strict=False, other_strict=strict)
        return self

    def bind_source_from(self,
                         target_object: Any,
                         target_name: str = 'source',
                         backward: Callable[[Any], Any] | None = None, *,
                         strict: bool | None = None,
                         ) -> Self:
        """Bind the source of this element from the target object's target_name property.

        The binding works one way only, from the target to this element.
        The update happens immediately and whenever a value changes.

        :param target_object: The object to bind from.
        :param target_name: The name of the property to bind from.
        :param backward: A function to apply to the value before applying it to this element (default: identity).
        :param strict: Whether to check (and raise) if the target object has the specified property (default: None,
            performs a check if the object is not a dictionary, *added in version 3.0.0*).
        """
        bind_from(self, 'source', target_object, target_name, backward, self_strict=False, other_strict=strict)
        return self

    def bind_source(self,
                    target_object: Any,
                    target_name: str = 'source', *,
                    forward: Callable[[Any], Any] | None = None,
                    backward: Callable[[Any], Any] | None = None,
                    strict: bool | None = None,
                    ) -> Self:
        """Bind the source of this element to the target object's target_name property.

        The binding works both ways, from this element to the target and from the target to this element.
        The update happens immediately and whenever a value changes.
        The backward binding takes precedence for the initial synchronization.

        :param target_object: The object to bind to.
        :param target_name: The name of the property to bind to.
        :param forward: A function to apply to the value before applying it to the target (default: identity).
        :param backward: A function to apply to the value before applying it to this element (default: identity).
        :param strict: Whether to check (and raise) if the target object has the specified property (default: None,
            performs a check if the object is not a dictionary, *added in version 3.0.0*).
        """
        bind(self, 'source', target_object, target_name,
             forward=forward, backward=backward,
             self_strict=False, other_strict=strict)
        return self

    def set_source(self, source: Any) -> None:
        """Set the source of this element.

        :param source: The new source.
        """
        self.source = source

    def _handle_source_change(self, source: Any) -> None:
        """Called when the source of this element changes.

        :param source: The new source.
        """
        self._set_props(source)

    def _set_props(self, source: Any) -> None:
        if is_file(source):
            self._temp_path = source  # prevent cleanup of _TempPath until source changes
            if self.auto_route:
                core.app.remove_route(self.auto_route)
            if self.SOURCE_IS_MEDIA_FILE:
                source = core.app.add_media_file(local_file=source)
            else:
                source = core.app.add_static_file(local_file=source)
            self.auto_route = source
        else:
            self._temp_path = None
        if isinstance(source, Path) and not source.exists():
            raise FileNotFoundError(f'File not found: {source}')
        self._props['src'] = source

    def _handle_delete(self) -> None:
        self._temp_path = None
        if self.auto_route:
            core.app.remove_route(self.auto_route)
        return super()._handle_delete()



from ..defaults import DEFAULT_PROP, resolve_defaults
from .mixins.color_elements import TextColorElement
from .mixins.name_element import NameElement


class Icon(NameElement, TextColorElement):

    @resolve_defaults
    def __init__(self,
                 name: str,
                 *,
                 size: str | None = DEFAULT_PROP | None,
                 color: str | None = DEFAULT_PROP | None,
                 ) -> None:
        """Icon

        This element is based on Quasar's `QIcon <https://quasar.dev/vue-components/icon>`_ component.

        `Here <https://fonts.google.com/icons?icon.set=Material+Icons>`_ is a reference of possible names.

        :param name: name of the icon (snake case, e.g. `add_circle`)
        :param size: size in CSS units, including unit name or standard size name (xs|sm|md|lg|xl), examples: 16px, 2rem
        :param color: icon color (either a Quasar, Tailwind, or CSS color or `None`, default: `None`)
        """
        super().__init__(tag='q-icon', name=name, text_color=color)

        self._props.set_optional('size', size)



from __future__ import annotations

import asyncio
from collections.abc import Callable
from typing import TYPE_CHECKING

from fastapi import Request
from starlette.routing import Match, Route

from . import core, json
from .context import context
from .elements.sub_pages import SubPages
from .functions.on import on

if TYPE_CHECKING:
    from .client import Client


class SubPagesRouter:

    def __init__(self, request: Request | None) -> None:
        on('sub_pages_open', lambda event: self._handle_open(event.args))
        on('sub_pages_navigate', lambda event: self._handle_navigate(event.args))

        if request is not None:
            forwarded_prefix = request.headers.get('X-Forwarded-Prefix', '')
            root = _normalize(request.scope.get('root_path', ''))
            combined = _normalize(forwarded_prefix or '') + _normalize(root or '')
            path = request.url.path
            for p in (combined, root):
                if p and (path == p or path.startswith(p + '/')):
                    path = path[len(p):] or '/'
                    break
            if request.url.query:
                path += '?' + request.url.query
            # NOTE: we do not use request.url.fragment because browsers do not send it to the server
            self.current_path = path
        else:
            self.current_path = '/'
        self.is_initial_request = True

        self._path_changed_handlers: list[Callable[[str], None]] = []

    def on_path_changed(self, handler: Callable[[str], None]) -> None:
        """Register a callback to be invoked when the path changes.

        :param handler: callback function that receives the new path as its argument
        """
        self._path_changed_handlers.append(handler)

    async def refresh(self) -> None:
        """Refresh the currently shown sub pages.

        This will clear and rebuild the current sub page as if navigating to it again.
        Useful when you want to update the page content based on changes in data or state.

        *Added in version 3.1.0*
        """
        for el in context.client.layout.descendants():
            if isinstance(el, SubPages):
                el._reset_match()  # pylint: disable=protected-access
        await self._handle_open(self.current_path)

    async def _handle_open(self, path: str) -> bool:
        self.current_path = path
        self.is_initial_request = False
        for callback in self._path_changed_handlers:
            callback(path)
        for child in context.client.layout.descendants():
            if isinstance(child, SubPages):
                child._show()  # pylint: disable=protected-access
        return await self._can_resolve_full_path(context.client)

    async def _handle_navigate(self, path: str) -> None:
        # NOTE: keep a reference to the client because _handle_open clears the slots so that context.client does not work anymore
        client = context.client
        await self._handle_open(path)
        if (
            not has_any_unresolved_path(client) or  # path is handled by `ui.sub_pages`
            not self._other_page_builder_matches_path(path, client)  # `ui.sub_pages` is still responsible
        ):
            current_path_string = json.dumps(self.current_path)
            client.run_javascript(f'''
                const fullPath = (window.path_prefix || '') + {current_path_string};
                if (window.location.pathname + window.location.search + window.location.hash !== fullPath) {{
                    history.pushState({{page: {current_path_string}}}, "", fullPath);
                }}
            ''')
        else:
            client.open(path, new_tab=False)

    def _other_page_builder_matches_path(self, path: str, client: Client) -> bool:
        """Check if there is any other matching page builder than the one for this client."""
        client_route = client.request.scope.get('route')
        client_func = getattr(client_route.endpoint, '__func__', client_route.endpoint) if client_route is not None else None

        other_routes = [route for route in core.app.routes if isinstance(route, Route)]
        for other_route in other_routes:
            other_func = getattr(other_route.endpoint, '__func__', other_route.endpoint)
            if (
                client_func is not None and
                getattr(client_func, '__name__', None) == getattr(other_func, '__name__', None) and
                getattr(client_func, '__module__', None) == getattr(other_func, '__module__', None) and
                getattr(client_func, '__qualname__', None) == getattr(other_func, '__qualname__', None)
            ):
                continue  # client route and other route point to the same page builder, so they don't count

            match, _ = other_route.matches({'type': 'http', 'path': path, 'method': 'GET'})
            if match == Match.FULL:
                return True

        return False

    @staticmethod
    async def _can_resolve_full_path(client: Client) -> bool:
        sub_pages_elements = [el for el in client.layout.descendants() if isinstance(el, SubPages)]
        if any(el._active_tasks for el in sub_pages_elements):  # pylint: disable=protected-access
            await asyncio.sleep(0)
            # NOTE: refresh the list to include newly created nested sub pages in async sub page builders after the event loop tick
            sub_pages_elements = [el for el in client.layout.descendants() if isinstance(el, SubPages)]
        for sub_pages in sub_pages_elements:
            if (
                sub_pages._match is not None and  # pylint: disable=protected-access
                sub_pages._match.remaining_path and  # pylint: disable=protected-access
                not any(isinstance(el, SubPages) for el in sub_pages.descendants())
            ):
                sub_pages._set_match(None)  # pylint: disable=protected-access
        return not has_any_unresolved_path(client, with_404_enabled_only=True)


def has_any_unresolved_path(client: Client, *, with_404_enabled_only: bool = False) -> bool:
    """Check if any sub_pages has an unresolved path.

    :param with_404_enabled_only: whether to only consider sub_pages with show_404 enabled
    """
    return any(
        sub_pages.has_404 and (sub_pages._404_enabled or not with_404_enabled_only)  # pylint: disable=protected-access
        for sub_pages in client.layout.descendants()
        if isinstance(sub_pages, SubPages)
    )


def _normalize(p: str) -> str:
    return p[:-1] if p and p != '/' and p.endswith('/') else p



from collections.abc import Callable
from typing import Any, cast

from typing_extensions import Self

from ...binding import BindableProperty, bind, bind_from, bind_to
from ...element import Element
from ...events import GenericEventArguments, Handler, ValueChangeEventArguments, handle_event


class ValueElement(Element):
    VALUE_PROP: str = 'model-value'
    '''Name of the prop that holds the value of the element'''

    LOOPBACK: bool | None = True
    '''Whether to set the new value directly on the client or after getting an update from the server.

    - ``True``: The value is updated by sending a change event to the server which responds with an update.
    - ``False``: The value is updated by setting the VALUE_PROP directly on the client.
    - ``None``: The value is updated automatically by the Vue element.
    '''

    value = BindableProperty(
        on_change=lambda sender, value: cast(Self, sender)._handle_value_change(value))  # pylint: disable=protected-access

    def __init__(self, *,
                 value: Any,
                 on_value_change: Handler[ValueChangeEventArguments] | None = None,
                 throttle: float = 0,
                 **kwargs: Any,
                 ) -> None:
        super().__init__(**kwargs)
        self._send_update_on_value_change = True
        self.set_value(value)
        self._props[self.VALUE_PROP] = self._value_to_model_value(value)
        self._props['loopback'] = self.LOOPBACK
        self._change_handlers: list[Handler[ValueChangeEventArguments]] = [on_value_change] if on_value_change else []

        def handle_change(e: GenericEventArguments) -> None:
            self._send_update_on_value_change = self.LOOPBACK is True
            self.set_value(self._event_args_to_value(e))
            self._send_update_on_value_change = True
        self.on(f'update:{self.VALUE_PROP}', handle_change, [None], throttle=throttle)

    def on_value_change(self, callback: Handler[ValueChangeEventArguments]) -> Self:
        """Add a callback to be invoked when the value changes."""
        self._change_handlers.append(callback)
        return self

    def bind_value_to(self,
                      target_object: Any,
                      target_name: str = 'value',
                      forward: Callable[[Any], Any] | None = None, *,
                      strict: bool | None = None,
                      ) -> Self:
        """Bind the value of this element to the target object's target_name property.

        The binding works one way only, from this element to the target.
        The update happens immediately and whenever a value changes.

        :param target_object: The object to bind to.
        :param target_name: The name of the property to bind to.
        :param forward: A function to apply to the value before applying it to the target (default: identity).
        :param strict: Whether to check (and raise) if the target object has the specified property (default: None,
            performs a check if the object is not a dictionary, *added in version 3.0.0*).
        """
        bind_to(self, 'value', target_object, target_name, forward, self_strict=False, other_strict=strict)
        return self

    def bind_value_from(self,
                        target_object: Any,
                        target_name: str = 'value',
                        backward: Callable[[Any], Any] | None = None, *,
                        strict: bool | None = None,
                        ) -> Self:
        """Bind the value of this element from the target object's target_name property.

        The binding works one way only, from the target to this element.
        The update happens immediately and whenever a value changes.

        :param target_object: The object to bind from.
        :param target_name: The name of the property to bind from.
        :param backward: A function to apply to the value before applying it to this element (default: identity).
        :param strict: Whether to check (and raise) if the target object has the specified property (default: None,
            performs a check if the object is not a dictionary, *added in version 3.0.0*).
        """
        bind_from(self, 'value', target_object, target_name, backward, self_strict=False, other_strict=strict)
        return self

    def bind_value(self,
                   target_object: Any,
                   target_name: str = 'value', *,
                   forward: Callable[[Any], Any] | None = None,
                   backward: Callable[[Any], Any] | None = None,
                   strict: bool | None = None,
                   ) -> Self:
        """Bind the value of this element to the target object's target_name property.

        The binding works both ways, from this element to the target and from the target to this element.
        The update happens immediately and whenever a value changes.
        The backward binding takes precedence for the initial synchronization.

        :param target_object: The object to bind to.
        :param target_name: The name of the property to bind to.
        :param forward: A function to apply to the value before applying it to the target (default: identity).
        :param backward: A function to apply to the value before applying it to this element (default: identity).
        :param strict: Whether to check (and raise) if the target object has the specified property (default: None,
            performs a check if the object is not a dictionary, *added in version 3.0.0*).
        """
        bind(self, 'value', target_object, target_name,
             forward=forward, backward=backward,
             self_strict=False, other_strict=strict)
        return self

    def set_value(self, value: Any) -> None:
        """Set the value of this element.

        :param value: The value to set.
        """
        self.value = value

    def _handle_value_change(self, value: Any) -> None:
        previous_value = self._props.get(self.VALUE_PROP)
        with self._props.suspend_updates():
            self._props[self.VALUE_PROP] = self._value_to_model_value(value)
        if self._send_update_on_value_change:
            self.update()
        args = ValueChangeEventArguments(sender=self, client=self.client,
                                         value=self._value_to_event_value(value),
                                         previous_value=self._value_to_event_value(previous_value))
        for handler in self._change_handlers:
            handle_event(handler, args)

    def _event_args_to_value(self, e: GenericEventArguments) -> Any:
        return e.args

    def _value_to_model_value(self, value: Any) -> Any:
        return value

    def _value_to_event_value(self, value: Any) -> Any:
        return value



from collections.abc import Iterator, MutableMapping
from typing import Any


class ReadOnlyDict(MutableMapping):

    def __init__(self, data: dict[Any, Any], write_error_message: str = 'Read-only dict') -> None:
        self._data: dict[Any, Any] = data
        self._write_error_message: str = write_error_message

    def __getitem__(self, item: Any) -> Any:
        return self._data[item]

    def __setitem__(self, key: Any, value: Any) -> None:
        raise TypeError(self._write_error_message)

    def __delitem__(self, key: Any) -> None:
        raise TypeError(self._write_error_message)

    def __iter__(self) -> Iterator:
        return iter(self._data)

    def __len__(self) -> int:
        return len(self._data)



from collections.abc import Callable
from typing import Any, cast

from typing_extensions import Self

from ...binding import BindableProperty, bind, bind_from, bind_to
from ...element import Element


class FilterElement(Element):
    FILTER_PROP = 'filter'
    filter = BindableProperty(
        on_change=lambda sender, filter: cast(Self, sender)._handle_filter_change(filter))  # pylint: disable=protected-access

    def __init__(self, *, filter: str | None = None, **kwargs: Any) -> None:  # pylint: disable=redefined-builtin
        super().__init__(**kwargs)
        self.filter = filter
        self._props[self.FILTER_PROP] = filter

    def bind_filter_to(self,
                       target_object: Any,
                       target_name: str = 'filter',
                       forward: Callable[[Any], Any] | None = None, *,
                       strict: bool | None = None,
                       ) -> Self:
        """Bind the filter of this element to the target object's target_name property.

        The binding works one way only, from this element to the target.
        The update happens immediately and whenever a value changes.

        :param target_object: The object to bind to.
        :param target_name: The name of the property to bind to.
        :param forward: A function to apply to the value before applying it to the target (default: identity).
        :param strict: Whether to check (and raise) if the target object has the specified property (default: None,
            performs a check if the object is not a dictionary, *added in version 3.0.0*).
        """
        bind_to(self, 'filter', target_object, target_name, forward, self_strict=False, other_strict=strict)
        return self

    def bind_filter_from(self,
                         target_object: Any,
                         target_name: str = 'filter',
                         backward: Callable[[Any], Any] | None = None, *,
                         strict: bool | None = None,
                         ) -> Self:
        """Bind the filter of this element from the target object's target_name property.

        The binding works one way only, from the target to this element.
        The update happens immediately and whenever a value changes.

        :param target_object: The object to bind from.
        :param target_name: The name of the property to bind from.
        :param backward: A function to apply to the value before applying it to this element (default: identity).
        :param strict: Whether to check (and raise) if the target object has the specified property (default: None,
            performs a check if the object is not a dictionary, *added in version 3.0.0*).
        """
        bind_from(self, 'filter', target_object, target_name, backward, self_strict=False, other_strict=strict)
        return self

    def bind_filter(self,
                    target_object: Any,
                    target_name: str = 'filter', *,
                    forward: Callable[[Any], Any] | None = None,
                    backward: Callable[[Any], Any] | None = None,
                    strict: bool | None = None,
                    ) -> Self:
        """Bind the filter of this element to the target object's target_name property.

        The binding works both ways, from this element to the target and from the target to this element.
        The update happens immediately and whenever a value changes.
        The backward binding takes precedence for the initial synchronization.

        :param target_object: The object to bind to.
        :param target_name: The name of the property to bind to.
        :param forward: A function to apply to the value before applying it to the target (default: identity).
        :param backward: A function to apply to the value before applying it to this element (default: identity).
        :param strict: Whether to check (and raise) if the target object has the specified property (default: None,
            performs a check if the object is not a dictionary, *added in version 3.0.0*).
        """
        bind(self, 'filter', target_object, target_name,
             forward=forward, backward=backward,
             self_strict=False, other_strict=strict)
        return self

    def set_filter(self, filter_: str) -> None:
        """Set the filter of this element.

        :param filter: The new filter.
        """
        self.filter = filter_

    def _handle_filter_change(self, filter_: str) -> None:
        """Called when the filter of this element changes.

        :param filter: The new filter.
        """
        self._props[self.FILTER_PROP] = filter_



from __future__ import annotations

import asyncio
import inspect
import weakref
from collections.abc import Awaitable, Callable
from contextlib import nullcontext
from dataclasses import dataclass
from typing import Any, ClassVar, Generic
from weakref import WeakSet

from typing_extensions import ParamSpec

from . import background_tasks, core, helpers
from .awaitable_response import AwaitableResponse
from .client import Client
from .context import context
from .slot import Slot

P = ParamSpec('P')


@dataclass(kw_only=True, slots=True)
class Callback(Generic[P]):
    func: Callable[P, Any] | Callable[[], Any]
    filepath: str
    line: int
    slot: weakref.ref[Slot] | None = None

    def run(self, *args: P.args, **kwargs: P.kwargs) -> Any:
        """Run the callback."""
        with (self.slot and self.slot()) or nullcontext():
            expect_args = helpers.expects_arguments(self.func)
            expect_args |= (
                isinstance(getattr(self.func, '__self__', None), Event) and
                getattr(self.func, '__name__', None) in {'emit', 'call'}
            )
            return self.func(*args, **kwargs) if expect_args else self.func()  # type: ignore[call-arg]

    async def await_result(self, result: Awaitable | AwaitableResponse | asyncio.Task) -> Any:
        """Await the result of the callback."""
        with (self.slot and self.slot()) or nullcontext():
            return await result


class Event(Generic[P]):
    instances: ClassVar[WeakSet[Event]] = WeakSet()

    def __init__(self) -> None:
        """Event

        Events are a powerful tool distribute information between different parts of your code,
        especially from long-living objects like data models to the short-living UI.

        Handlers can be synchronous or asynchronous.
        They can also take arguments if the event contains arguments.

        *Added in version 3.0.0*
        """
        self.callbacks: list[Callback[P]] = []
        self.instances.add(self)

    def subscribe(self, callback: Callable[P, Any] | Callable[[], Any], *,
                  unsubscribe_on_delete: bool | None = None) -> None:
        """Subscribe to the event.

        The ``unsubscribe_on_delete`` can be used to explicitly define
        whether the callback should be automatically unsubscribed when the current client is deleted.
        By default, the callback is automatically unsubscribed if subscribed from within a UI context
        to prevent memory leaks.

        :param callback: the callback which will be called when the event is fired
        :param unsubscribe_on_delete: whether to unsubscribe the callback when the current client is deleted
            (default: ``None`` meaning the callback is automatically unsubscribed if subscribed from within a UI context)
        """
        frame = inspect.currentframe()
        assert frame is not None
        frame = frame.f_back
        assert frame is not None
        callback_ = Callback[P](func=callback, filepath=frame.f_code.co_filename, line=frame.f_lineno)
        client: Client | None = None
        if Slot.get_stack():  # NOTE: additional check before accessing `context.slot` which would enter script mode
            callback_.slot = weakref.ref(context.slot)
            client = context.client
        if callback_.slot is None and unsubscribe_on_delete is True:
            raise RuntimeError('Calling `subscribe` with `unsubscribe_on_delete=True` outside of a UI context '
                               'is not supported.')
        if client is not None and unsubscribe_on_delete is not False and not core.is_script_mode_preflight():
            client.on_delete(lambda: self.unsubscribe(callback))
        self.callbacks.append(callback_)

    def unsubscribe(self, callback: Callable[P, Any] | Callable[[], Any]) -> None:
        """Unsubscribe a callback from the event.

        :param callback: the callback to unsubscribe from the event
        """
        self.callbacks[:] = [c for c in self.callbacks if c.func != callback]

    def emit(self, *args: P.args, **kwargs: P.kwargs) -> None:
        """Fire the event without waiting for the subscribed callbacks to complete."""
        for callback in self.callbacks:
            _invoke_and_forget(callback, *args, **kwargs)

    async def call(self, *args: P.args, **kwargs: P.kwargs) -> None:
        """Fire the event and wait asynchronously until all subscribed callbacks are completed."""
        await asyncio.gather(*[_invoke_and_await(callback, *args, **kwargs) for callback in self.callbacks])

    async def emitted(self, timeout: float | None = None) -> Any:
        """Wait for an event to be fired and return its arguments.

        :param timeout: the maximum time to wait for the event to be fired (default: ``None`` meaning no timeout)
        """
        future: asyncio.Future[Any] = asyncio.Future()

        def callback(*args: P.args, **kwargs: P.kwargs) -> None:  # pylint: disable=unused-argument
            if not future.done():
                future.set_result(args[0] if len(args) == 1 else args if args else None)

        self.subscribe(callback)
        try:
            return await asyncio.wait_for(future, timeout)
        except TimeoutError as error:
            raise TimeoutError(f'Timed out waiting for event after {timeout} seconds') from error
        finally:
            self.unsubscribe(callback)

    def __await__(self):
        return self.emitted().__await__()


def _invoke_and_forget(callback: Callback[P], *args: P.args, **kwargs: P.kwargs) -> Any:
    try:
        result = callback.run(*args, **kwargs)
        if _should_await(result):
            if core.loop and core.loop.is_running():
                background_tasks.create(callback.await_result(result), name=f'{callback.filepath}:{callback.line}')
            else:
                core.app.on_startup(callback.await_result(result))
    except Exception as e:
        core.app.handle_exception(e)


async def _invoke_and_await(callback: Callback[P], *args: P.args, **kwargs: P.kwargs) -> Any:
    result = callback.run(*args, **kwargs)
    if _should_await(result):
        result = await callback.await_result(result)
    return result


def _should_await(result: Any) -> bool:
    """Determine if a result should be awaited.

    Note: We want to await an awaitable result even if the handler is not a coroutine (like a lambda statement).
    """
    return isinstance(result, Awaitable) and not isinstance(result, AwaitableResponse) and not isinstance(result, asyncio.Task)


def reset() -> None:
    """Reset the event system. (Useful for testing.)"""
    for event in Event.instances:
        event.callbacks.clear()
    Event.instances.clear()



from typing import Any

from ..element import Element
from .label import Label


class Log(Element, component='log.js', default_classes='nicegui-log'):

    def __init__(self, max_lines: int | None = None) -> None:
        """Log View

        Create a log view that allows to add new lines without re-transmitting the whole history to the client.

        :param max_lines: maximum number of lines before dropping oldest ones (default: `None`)
        """
        super().__init__()
        self.max_lines = max_lines

    def push(self, line: Any, *,
             classes: str | None = None,
             style: str | None = None,
             props: str | None = None) -> None:
        """Add a new line to the log.

        :param line: the line to add (can contain line breaks)
        :param classes: classes to apply to the line (*added in version 2.18.0*)
        :param style: style to apply to the line (*added in version 2.18.0*)
        :param props: props to apply to the line (*added in version 2.18.0*)
        """
        for text in str(line).splitlines():
            with self:
                label = Label(text)
                if classes is not None:
                    label.classes(replace=classes)
                if style is not None:
                    label.style(replace=style)
                if props is not None:
                    label.props.clear()
                    label.props(props)
        while self.max_lines is not None and len(self.default_slot.children) > self.max_lines:
            self.remove(0)



import uuid
from collections.abc import Callable, Sequence
from dataclasses import dataclass, field
from typing import Any

from fastapi import Request


@dataclass(kw_only=True, slots=True)
class EventListener:
    id: str = field(init=False)
    element_id: int
    type: str
    args: Sequence[Sequence[str] | None]
    handler: Callable | None
    js_handler: str | None
    throttle: float
    leading_events: bool
    trailing_events: bool
    request: Request | None

    def __post_init__(self) -> None:
        self.id = str(uuid.uuid4())

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of the event listener."""
        words = self.type.split('.')
        type_ = words.pop(0)
        specials = [w for w in words if w in {'capture', 'once', 'passive'}]
        modifiers = [w for w in words if w in {'stop', 'prevent', 'self', 'ctrl', 'shift', 'alt', 'meta'}]
        keys = [w for w in words if w not in specials + modifiers]
        return {
            'listener_id': self.id,
            'type': type_,
            'specials': specials,
            'modifiers': modifiers,
            'keys': keys,
            'args': self.args,
            'throttle': self.throttle,
            'leading_events': self.leading_events,
            'trailing_events': self.trailing_events,
            'js_handler': self.js_handler,
        }



from typing_extensions import Self

from ..defaults import DEFAULT_PROP, resolve_defaults
from ..events import ClickEventArguments, Handler, ValueChangeEventArguments, handle_event
from .mixins.color_elements import BackgroundColorElement, TextColorElement
from .mixins.disableable_element import DisableableElement
from .mixins.icon_element import IconElement
from .mixins.selectable_element import SelectableElement
from .mixins.text_element import TextElement
from .mixins.value_element import ValueElement


class Chip(IconElement, ValueElement, TextElement, BackgroundColorElement, TextColorElement, DisableableElement, SelectableElement):
    TEXT_COLOR_PROP = 'text-color'

    @resolve_defaults
    def __init__(self,
                 text: str = '',
                 *,
                 icon: str | None = None,
                 color: str | None = DEFAULT_PROP | 'primary',
                 text_color: str | None = DEFAULT_PROP | None,
                 on_click: Handler[ClickEventArguments] | None = None,
                 selectable: bool = DEFAULT_PROP | False,
                 selected: bool = DEFAULT_PROP | False,
                 on_selection_change: Handler[ValueChangeEventArguments] | None = None,
                 removable: bool = DEFAULT_PROP | False,
                 on_value_change: Handler[ValueChangeEventArguments] | None = None,
                 ) -> None:
        """Chip

        A chip element wrapping Quasar's `QChip <https://quasar.dev/vue-components/chip>`_ component.
        It can be clickable, selectable and removable.

        :param text: the initial value of the text field (default: "")
        :param icon: the name of an icon to be displayed on the chip (default: `None`)
        :param color: the color name for component (either a Quasar, Tailwind, or CSS color or `None`, default: "primary")
        :param text_color: text color (either a Quasar, Tailwind, or CSS color or `None`, default: `None`)
        :param on_click: callback which is invoked when chip is clicked. Makes the chip clickable if set
        :param selectable: whether the chip is selectable (default: `False`)
        :param selected: whether the chip is selected (default: `False`)
        :param on_selection_change: callback which is invoked when the chip's selection state is changed
        :param removable: whether the chip is removable. Shows a small "x" button if True (default: `False`)
        :param on_value_change: callback which is invoked when the chip is removed or unremoved
        """
        super().__init__(tag='q-chip', value=True, on_value_change=on_value_change,
                         icon=icon, text=text, text_color=text_color, background_color=color,
                         selectable=selectable, selected=selected, on_selection_change=on_selection_change)

        self._props['removable'] = removable

        if on_click:
            self.on_click(on_click)

    def on_click(self, callback: Handler[ClickEventArguments]) -> Self:
        """Add a callback to be invoked when the chip is clicked."""
        self._props['clickable'] = True
        self.on('click', lambda _: handle_event(callback, ClickEventArguments(sender=self, client=self.client)), [])
        return self



from __future__ import annotations

import inspect
import math
from pathlib import Path
from typing import TYPE_CHECKING, Any

from ... import helpers, optional_features
from ..mixins.value_element import ValueElement

optional_features.try_register('anywidget')
if TYPE_CHECKING:
    import anywidget

UNDEFINED = object()


class AnyWidget(ValueElement, component='anywidget.js', dependencies=['lib/widget.js']):
    VALUE_PROP: str = 'traits'

    def __init__(self, widget: anywidget.AnyWidget, *, throttle: float = 0) -> None:
        """AnyWidget

        `anywidget <https://anywidget.dev/en/getting-started/>`_ is a library that allows you to
        embed arbitrary JavaScript widgets in a cross-frontend friendly manner.

        There are many publicly available examples of anywidget widgets
        in the `anywidget gallery <https://try.anywidget.dev/>`_, including
        `altair.JupyterChart <https://altair-viz.github.io/user_guide/interactions/jupyter_chart.html>`_,
        and `quak <https://github.com/manzt/quak>`_.

        Implementation: The ``nicegui.anywidget`` element takes an ``AnyWidget`` and observes all ``sync=True`` traits
        of the widget, trigger JS updates when the traits change.
        Conversely, changes on the frontend will be synced back to the widget,
        using ``ValueElement``'s handling to listen to changes on ``traits``.

        *Added in version 3.5.0*

        :param widget: the ``anywidget.AnyWidget`` to wrap
        :param throttle: minimum time (in seconds) between widget updates to Python (default: 0.0)
        """
        self._widget = widget
        self._traits = widget.traits(sync=True)
        super().__init__(value=widget.get_state(self._traits), throttle=throttle)
        self._props['esm_content'] = _get_attribute(widget, '_esm')
        self._props['css_content'] = _get_attribute(widget, '_css')
        self._state_lock: dict | None = None  # only used while handling a value change from the client

        def observe_change(change) -> None:
            """Observe a trait change and update the frontend (but avoid echoing same values back to the client)."""
            name = change['name']
            new = change['new']
            if self._state_lock is None:
                # we're not handling a value change from the client, so we send an update to the client
                self.run_method('update_trait', name, new)
            elif not _equal(self._state_lock.get(name, UNDEFINED), new):
                # an observer changed a trait to a new value, so we update the lock and send an update to the client
                self._state_lock[name] = new
                self.run_method('update_trait', name, new)

        widget.observe(observe_change, self._traits)

    def _handle_value_change(self, value: Any) -> None:
        """Update the widget's state when the value changes from frontend"""
        self._state_lock = value
        try:
            super()._handle_value_change(value)
            state = self._widget.get_state(self._traits)
            for key, value_ in value.items():
                if state[key] != value_:
                    setattr(self._widget, key, value_)
        finally:
            self._state_lock = None


def _get_attribute(obj: object, name: str) -> str:
    """Extract the attribute's content, reading if it is a path to a file."""
    content = getattr(obj, name, '')
    if callable(content) and not inspect.isclass(content):  # content is a property function
        content = content()
    assert isinstance(content, (str, Path)), f'Attribute {name} is not a string or Path'
    if helpers.is_file(content):
        content = Path(content).read_text(encoding='utf8')
    assert isinstance(content, str), f'Attribute {name} is a Path but does not exist'
    return content


def _equal(a: Any, b: Any) -> bool:
    """Check if two values are equal, considering NaN as equal."""
    if a == b:
        return True
    try:
        return math.isnan(a) and math.isnan(b)
    except TypeError:
        return False



from typing import Any, Literal

from typing_extensions import Self

from .pyplot import Pyplot


class LinePlot(Pyplot):

    def __init__(self, *,
                 n: int = 1,
                 limit: int = 100,
                 update_every: int = 1,
                 close: bool = True,
                 **kwargs: Any,
                 ) -> None:
        """Line Plot

        Create a line plot using pyplot.
        The `push` method provides live updating when utilized in combination with `ui.timer`.

        :param n: number of lines
        :param limit: maximum number of datapoints per line (new points will displace the oldest)
        :param update_every: update plot only after pushing new data multiple times to save CPU and bandwidth
        :param close: whether the figure should be closed after exiting the context; set to `False` if you want to update it later (default: `True`)
        :param kwargs: arguments like `figsize` which should be passed to `pyplot.figure <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.figure.html>`_
        """
        super().__init__(close=close, **kwargs)

        self.x: list[float] = []
        self.Y: list[list[float]] = [[] for _ in range(n)]
        self.lines = [self.fig.gca().plot([], [])[0] for _ in range(n)]
        self.slice = slice(0 if limit is None else -limit, None)
        self.update_every = update_every
        self.push_counter = 0

    def with_legend(self, titles: list[str], **kwargs: Any):
        """Add a legend to the plot.

        :param titles: list of titles for the lines
        :param kwargs: additional arguments which should be passed to `pyplot.legend <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.legend.html>`_
        """
        self.fig.gca().legend(titles, **kwargs)
        self._convert_to_html()
        return self

    def push(self,
             x: list[float],
             Y: list[list[float]],
             *,
             x_limits: None | Literal['auto'] | tuple[float, float] = 'auto',
             y_limits: None | Literal['auto'] | tuple[float, float] = 'auto',
             ) -> None:
        """Push new data to the plot.

        :param x: list of x values
        :param Y: list of lists of y values (one list per line)
        :param x_limits: new x limits (tuple of floats, or "auto" to fit the data points, or ``None`` to leave unchanged, *added in version 2.10.0*)
        :param y_limits: new y limits (tuple of floats, or "auto" to fit the data points, or ``None`` to leave unchanged, *added in version 2.10.0*)
        """
        self.push_counter += 1

        self.x = [*self.x, *x][self.slice]
        for i in range(len(self.lines)):
            self.Y[i] = [*self.Y[i], *Y[i]][self.slice]

        if self.push_counter % self.update_every != 0:
            return

        for i, line in enumerate(self.lines):
            line.set_xdata(self.x)
            line.set_ydata(self.Y[i])

        if isinstance(x_limits, tuple):
            self.fig.gca().set_xlim(*x_limits)
        elif x_limits == 'auto':
            min_x = min(self.x)
            max_x = max(self.x)
            if min_x != max_x:
                pad_x = 0.01 * (max_x - min_x)
                self.fig.gca().set_xlim(min_x - pad_x, max_x + pad_x)

        if isinstance(y_limits, tuple):
            self.fig.gca().set_ylim(*y_limits)
        elif y_limits == 'auto':
            flat_y = [y_i for y in self.Y for y_i in y]
            min_y = min(flat_y)
            max_y = max(flat_y)
            if min_y != max_y:
                pad_y = 0.01 * (max_y - min_y)
                self.fig.gca().set_ylim(min_y - pad_y, max_y + pad_y)

        self._convert_to_html()

    def clear(self) -> Self:
        """Clear the line plot."""
        super().clear()
        self.x.clear()
        for y in self.Y:
            y.clear()
        for line in self.lines:
            line.set_data([], [])
        self._convert_to_html()
        return self



import abc

from .. import observables


class PersistentDict(observables.ObservableDict, abc.ABC):

    @abc.abstractmethod
    async def initialize(self) -> None:
        """Load initial data from the persistence layer."""

    @abc.abstractmethod
    def initialize_sync(self) -> None:
        """Load initial data from the persistence layer in a synchronous context."""

    async def close(self) -> None:
        """Clean up the persistence layer."""



from collections.abc import Callable
from typing import TYPE_CHECKING, Any, cast

from typing_extensions import Self

from ...binding import BindableProperty, bind, bind_from, bind_to

if TYPE_CHECKING:
    from ...element import Element


class Visibility:
    visible = BindableProperty(
        on_change=lambda sender, visible: cast(Self, sender)._handle_visibility_change(visible))  # pylint: disable=protected-access

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.visible = True
        self.ignores_events_when_hidden = True

    @property
    def is_ignoring_events(self) -> bool:
        """Return whether the element is currently ignoring events."""
        return not self.visible and self.ignores_events_when_hidden

    def bind_visibility_to(self,
                           target_object: Any,
                           target_name: str = 'visible',
                           forward: Callable[[Any], Any] | None = None, *,
                           strict: bool | None = None,
                           ) -> Self:
        """Bind the visibility of this element to the target object's target_name property.

        The binding works one way only, from this element to the target.
        The update happens immediately and whenever a value changes.

        :param target_object: The object to bind to.
        :param target_name: The name of the property to bind to.
        :param forward: A function to apply to the value before applying it to the target (default: identity).
        :param strict: Whether to check (and raise) if the target object has the specified property (default: None,
            performs a check if the object is not a dictionary, *added in version 3.0.0*).
        """
        bind_to(self, 'visible', target_object, target_name, forward, self_strict=False, other_strict=strict)
        return self

    def bind_visibility_from(self,
                             target_object: Any,
                             target_name: str = 'visible',
                             backward: Callable[[Any], Any] | None = None, *,
                             value: Any = None,
                             strict: bool | None = None,
                             ) -> Self:
        """Bind the visibility of this element from the target object's target_name property.

        The binding works one way only, from the target to this element.
        The update happens immediately and whenever a value changes.

        :param target_object: The object to bind from.
        :param target_name: The name of the property to bind from.
        :param backward: A function to apply to the value before applying it to this element (default: identity).
        :param value: If specified, the element will be visible only when the target value is equal to this value.
        :param strict: Whether to check (and raise) if the target object has the specified property (default: None,
            performs a check if the object is not a dictionary, *added in version 3.0.0*).
        """
        if value is not None:
            def backward(x):  # pylint: disable=function-redefined
                return x == value
        bind_from(self, 'visible', target_object, target_name, backward, self_strict=False, other_strict=strict)
        return self

    def bind_visibility(self,
                        target_object: Any,
                        target_name: str = 'visible', *,
                        forward: Callable[[Any], Any] | None = None,
                        backward: Callable[[Any], Any] | None = None,
                        value: Any = None,
                        strict: bool | None = None,
                        ) -> Self:
        """Bind the visibility of this element to the target object's target_name property.

        The binding works both ways, from this element to the target and from the target to this element.
        The update happens immediately and whenever a value changes.
        The backward binding takes precedence for the initial synchronization.

        :param target_object: The object to bind to.
        :param target_name: The name of the property to bind to.
        :param forward: A function to apply to the value before applying it to the target (default: identity).
        :param backward: A function to apply to the value before applying it to this element (default: identity).
        :param value: If specified, the element will be visible only when the target value is equal to this value.
        :param strict: Whether to check (and raise) if the target object has the specified property (default: None,
            performs a check if the object is not a dictionary, *added in version 3.0.0*).
        """
        if value is not None:
            def backward(x):  # pylint: disable=function-redefined
                return x == value
        bind(self, 'visible', target_object, target_name,
             forward=forward, backward=backward,
             self_strict=False, other_strict=strict)
        return self

    def set_visibility(self, visible: bool) -> None:
        """Set the visibility of this element.

        :param visible: Whether the element should be visible.
        """
        self.visible = visible

    def _handle_visibility_change(self, visible: str) -> None:
        """Called when the visibility of this element changes.

        :param visible: Whether the element should be visible.
        """
        element: Element = cast('Element', self)
        classes = element.classes  # pylint: disable=no-member
        if visible and 'hidden' in classes:
            classes.remove('hidden')
        if not visible and 'hidden' not in classes:
            classes.append('hidden')



import asyncio
from pathlib import Path
from typing import Any, cast

from typing_extensions import Self

from ... import binding
from ...awaitable_response import AwaitableResponse, NullResponse
from ...defaults import DEFAULT_PROP, resolve_defaults
from ...element import Element
from ...events import GenericEventArguments
from .leaflet_layer import Layer


class Leaflet(Element, component='leaflet.js', esm={'nicegui-leaflet': 'dist'}, default_classes='nicegui-leaflet'):
    # pylint: disable=import-outside-toplevel
    from .leaflet_layers import GenericLayer as generic_layer
    from .leaflet_layers import ImageOverlay as image_overlay
    from .leaflet_layers import Marker as marker
    from .leaflet_layers import TileLayer as tile_layer
    from .leaflet_layers import VideoOverlay as video_overlay
    from .leaflet_layers import WmsLayer as wms_layer

    center = binding.BindableProperty(lambda sender, value: cast(Leaflet, sender).set_center(value))
    zoom = binding.BindableProperty(lambda sender, value: cast(Leaflet, sender).set_zoom(value))

    @resolve_defaults
    def __init__(self,
                 center: tuple[float, float] = DEFAULT_PROP | (0.0, 0.0),
                 zoom: int = DEFAULT_PROP | 13,
                 *,
                 options: dict = DEFAULT_PROP | {},
                 draw_control: bool | dict = DEFAULT_PROP | False,
                 hide_drawn_items: bool = DEFAULT_PROP | False,
                 additional_resources: list[str] | None = DEFAULT_PROP | None,
                 ) -> None:
        """Leaflet map

        This element is a wrapper around the `Leaflet <https://leafletjs.com/>`_ JavaScript library.

        :param center: initial center location of the map (latitude/longitude, default: (0.0, 0.0))
        :param zoom: initial zoom level of the map (default: 13)
        :param draw_control: whether to show the draw toolbar (default: False)
        :param options: additional options passed to the Leaflet map (default: {})
        :param hide_drawn_items: whether to hide drawn items on the map (default: False, *added in version 2.0.0*)
        :param additional_resources: additional resources like CSS or JS files to load (default: None, *added in version 2.11.0*)
        """
        super().__init__()
        self.add_resource(Path(__file__).parent / 'dist')

        self.layers: list[Layer] = []
        self.is_initialized = False

        # read-write public API
        self.center = center
        self.zoom = zoom

        # internal state, mutates client state via _to_dict
        self._props['center'] = center
        self._props['zoom'] = zoom

        # client state
        self._client_center = center
        self._client_zoom = zoom

        self._props['options'] = {**options}
        self._props['draw-control'] = draw_control
        self._props['hide-drawn-items'] = hide_drawn_items
        self._props['additional-resources'] = additional_resources or []

        self._props.add_rename('draw_control', 'draw-control')  # DEPRECATED: remove in NiceGUI 4.0
        self._props.add_rename('hide_drawn_items', 'hide-drawn-items')  # DEPRECATED: remove in NiceGUI 4.0
        self._props.add_rename('additional_resources', 'additional-resources')  # DEPRECATED: remove in NiceGUI 4.0

        self.on('init', self._handle_init)
        self.on('map-moveend', self._handle_move_or_zoom_end)
        self.on('map-zoomend', self._handle_move_or_zoom_end)

        self.tile_layer(
            url_template=r'https://{s}.tile.osm.org/{z}/{x}/{y}.png',
            options={'attribution': '&copy; <a href="https://openstreetmap.org">OpenStreetMap</a> contributors'},
        )

        self._send_update_on_value_change = True

    def __enter__(self) -> Self:
        Layer.current_leaflet = self
        return super().__enter__()

    def __getattribute__(self, name: str) -> Any:
        attribute = super().__getattribute__(name)
        if isinstance(attribute, type) and issubclass(attribute, Layer):
            Layer.current_leaflet = self
        return attribute

    def _handle_init(self) -> None:
        self.is_initialized = True
        for layer in self.layers:
            self.run_method('add_layer', layer.to_dict(), layer.id)

    async def initialized(self) -> None:
        """Wait until the map is initialized."""
        event = asyncio.Event()
        self.on('init', event.set, [])
        await self.client.connected()
        await event.wait()

    def _handle_move_or_zoom_end(self, e: GenericEventArguments) -> None:
        self._send_update_on_value_change = False
        self.center = self._client_center = e.args['center']
        self.zoom = self._client_zoom = e.args['zoom']
        self._send_update_on_value_change = True

    def run_method(self, name: str, *args: Any, timeout: float = 1) -> AwaitableResponse:
        if not self.is_initialized:
            return NullResponse()
        return super().run_method(name, *args, timeout=timeout)

    def set_center(self, center: tuple[float, float]) -> None:
        """Set the center location of the map."""
        if self._props['center'] == center:
            return
        self._props['center'] = center

    def set_zoom(self, zoom: int) -> None:
        """Set the zoom level of the map."""
        if self._props['zoom'] == zoom:
            return
        self._props['zoom'] = zoom

    def remove_layer(self, layer: Layer) -> None:
        """Remove a layer from the map."""
        self.layers.remove(layer)
        self.run_method('remove_layer', layer.id)

    def clear_layers(self) -> None:
        """Remove all layers from the map."""
        self.layers.clear()
        self.run_method('clear_layers')

    def run_map_method(self, name: str, *args, timeout: float = 1) -> AwaitableResponse:
        """Run a method of the Leaflet map instance.

        Refer to the `Leaflet documentation <https://leafletjs.com/reference.html#map-methods-for-modifying-map-state>`_ for a list of methods.

        If the function is awaited, the result of the method call is returned.
        Otherwise, the method is executed without waiting for a response.

        :param name: name of the method (a prefix ":" indicates that the arguments are JavaScript expressions)
        :param args: arguments to pass to the method
        :param timeout: timeout in seconds (default: 1 second)

        :return: AwaitableResponse that can be awaited to get the result of the method call
        """
        return self.run_method('run_map_method', name, *args, timeout=timeout)

    def run_layer_method(self, layer_id: str, name: str, *args, timeout: float = 1) -> AwaitableResponse:
        """Run a method of a Leaflet layer.

        If the function is awaited, the result of the method call is returned.
        Otherwise, the method is executed without waiting for a response.

        :param layer_id: ID of the layer
        :param name: name of the method (a prefix ":" indicates that the arguments are JavaScript expressions)
        :param args: arguments to pass to the method
        :param timeout: timeout in seconds (default: 1 second)

        :return: AwaitableResponse that can be awaited to get the result of the method call
        """
        return self.run_method('run_layer_method', layer_id, name, *args, timeout=timeout)

    def _handle_delete(self) -> None:
        binding.remove(self.layers)
        super()._handle_delete()

    def _to_dict(self):
        if self._send_update_on_value_change and (
            self._props['center'] != self._client_center or
            self._props['zoom'] != self._client_zoom
        ):
            self.run_map_method('setView', self._props['center'], self._props['zoom'])
        return super()._to_dict()



import importlib.util
from decimal import Decimal
from typing import Any

# pylint: disable=no-member
import orjson
from fastapi.responses import JSONResponse

try:
    HAS_NUMPY = importlib.util.find_spec('numpy') is not None
except (ModuleNotFoundError, ValueError):
    HAS_NUMPY = False

ORJSON_OPTS = orjson.OPT_SERIALIZE_NUMPY | orjson.OPT_NON_STR_KEYS


def dumps(obj: Any,
          sort_keys: bool = False,
          separators: tuple[str, str] | None = None, *,
          indent: bool = False) -> str:
    """Serializes a Python object to a JSON-encoded string.

    By default, this function supports serializing NumPy arrays, which Python's json module does not.

    Uses package `orjson` internally.
    """
    # note that parameters `sort_keys` and `separators` are required by AsyncServer's
    # internal calls, which match Python's default `json.dumps` API.
    assert separators is None or separators == (',', ':'), \
        'NiceGUI JSON serializer only supports Python''s default ' +\
        f'JSON separators "," and ":", but got {separators} instead.'

    opts = ORJSON_OPTS

    # flag for sorting by object keys
    if sort_keys:
        opts |= orjson.OPT_SORT_KEYS

    # flag for pretty-printing with indentation
    if indent:
        opts |= orjson.OPT_INDENT_2

    return orjson.dumps(obj, option=opts, default=_orjson_converter).decode('utf-8')


def loads(value: str) -> Any:
    """Deserialize a JSON-encoded string to a corresponding Python object/value.

    Uses package `orjson` internally.
    """
    return orjson.loads(value)


def _orjson_converter(obj):
    """Custom serializer/converter, e.g. for NumPy object arrays."""
    if HAS_NUMPY:
        import numpy as np  # pylint: disable=import-outside-toplevel
        if isinstance(obj, np.ndarray) and obj.dtype == np.object_:
            return obj.tolist()
    if isinstance(obj, Decimal):
        return float(obj)
    raise TypeError(f'Object of type {obj.__class__.__name__} is not JSON serializable')


class NiceGUIJSONResponse(JSONResponse):
    """FastAPI response class to support our custom json serializer implementation.

    Uses package `orjson` internally.
    """

    def render(self, content: Any) -> bytes:
        return orjson.dumps(content, option=ORJSON_OPTS, default=_orjson_converter)



from .helpers import warn_once

# DEPRECATED: This module will be removed in NiceGUI 4.0


def __getattr__(name: str) -> dict:
    if name == 'KWONLY_SLOTS':
        warn_once('nicegui.dataclasses.KWONLY_SLOTS is deprecated, use kw_only=True, slots=True directly')
        return {'kw_only': True, 'slots': True}
    raise AttributeError(f'module {__name__!r} has no attribute {name!r}')



from __future__ import annotations

from typing import TYPE_CHECKING

from .. import optional_features
from .anywidget import AnyWidget

optional_features.try_register('altair')
if TYPE_CHECKING:
    import altair


class Altair(AnyWidget):

    def __init__(self, chart: altair.TopLevelSpec | altair.JupyterChart, *, throttle: float = 0) -> None:
        """Altair Chart

        Wrap an Altair chart in NiceGUI via anywidget.

        Refer to the `altair documentation <https://altair-viz.github.io/user_guide/interactions/jupyter_chart.html#accessing-variable-params>`_
        for more information about synchronizing Altair parameters with Python.

        *Added in version 3.5.0*

        :param chart: the chart to wrap
        :param throttle: minimum time (in seconds) between widget updates to Python (default: 0.0)
        """
        import altair  # pylint: disable=import-outside-toplevel

        if isinstance(chart, altair.TopLevelSpec):
            chart = altair.JupyterChart(chart)

        super().__init__(chart, throttle=throttle)  # type: ignore[arg-type]



from dataclasses import dataclass, field
from typing import Any

from typing_extensions import Self

from .leaflet_layer import Layer


@dataclass(kw_only=True, slots=True)
class GenericLayer(Layer):
    name: str
    args: list[Any] = field(default_factory=list)

    def to_dict(self) -> dict:
        return {
            'type': self.name,
            'args': self.args,
        }


@dataclass(kw_only=True, slots=True)
class TileLayer(Layer):
    url_template: str
    options: dict = field(default_factory=dict)

    def to_dict(self) -> dict:
        return {
            'type': 'tileLayer',
            'args': [self.url_template, self.options],
        }


@dataclass(kw_only=True, slots=True)
class WmsLayer(Layer):
    url_template: str
    options: dict = field(default_factory=dict)

    def to_dict(self) -> dict:
        return {
            'type': 'tileLayer.wms',
            'args': [self.url_template, self.options],
        }


@dataclass(kw_only=True, slots=True)
class ImageOverlay(Layer):
    url: str
    bounds: list[list[float]]
    options: dict = field(default_factory=dict)

    def to_dict(self) -> dict:
        return {
            'type': 'imageOverlay',
            'args': [self.url, self.bounds, self.options],
        }


@dataclass(kw_only=True, slots=True)
class VideoOverlay(Layer):
    url: str | list[str]
    bounds: list[list[float]]
    options: dict = field(default_factory=dict)

    def to_dict(self) -> dict:
        return {
            'type': 'videoOverlay',
            'args': [self.url, self.bounds, self.options],
        }


@dataclass(kw_only=True, slots=True)
class Marker(Layer):
    latlng: tuple[float, float]
    options: dict = field(default_factory=dict)

    def to_dict(self) -> dict:
        return {
            'type': 'marker',
            'args': [{'lat': self.latlng[0], 'lng': self.latlng[1]}, self.options],
        }

    def draggable(self, value: bool = True) -> Self:
        """Make the marker draggable."""
        self.options['draggable'] = value
        return self

    def move(self, lat: float, lng: float) -> None:
        """Move the marker to a new position.

        :param lat: latitude
        :param lng: longitude
        """
        self.latlng = (lat, lng)
        self.run_method('setLatLng', (lat, lng))



from __future__ import annotations

import inspect
import types
from collections.abc import Callable
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Union, get_args, get_origin

from starlette.datastructures import QueryParams

if TYPE_CHECKING:
    from .elements.sub_pages import SubPages


@dataclass(kw_only=True, slots=True)
class RouteMatch:
    """Contains details about a matched route including path parameters and query data."""
    path: str
    '''The sub-path that actually matched (e.g., "/user/123").'''
    pattern: str
    '''The original route pattern (e.g., "/user/{id}").'''
    builder: Callable
    '''The function to call to build the page.'''
    parameters: dict[str, str]
    '''The extracted parameters (name -> value) from the path (e.g., ``{'id': '123'}``).'''
    query_params: QueryParams
    '''The query parameters from the URL.'''
    fragment: str
    '''The URL fragment (e.g., "section" from "#section").'''
    remaining_path: str = ''
    '''The remaining path after the matched path.'''

    @property
    def full_url(self) -> str:
        """Get the complete URL including path and query parameters, but excluding fragment.

        Fragments should not trigger page rebuilds, only scrolling behavior.
        """
        url = self.path
        if self.query_params:
            url += '?' + str(self.query_params)
        return url

    def __repr__(self) -> str:
        return f'{self.path=}, {self.pattern=}, builder={self.builder.__name__}, {self.parameters=}, {self.query_params=}, {self.fragment=}, {self.remaining_path=}'


@dataclass(kw_only=True, slots=True)
class PageArguments:
    """Provides unified access to route data including path parameters and query parameters.

    Automatically injected into sub-page builder functions when the parameter is annotated with ``PageArguments``.
    """
    path: str
    '''Path from the request.'''
    frame: SubPages
    '''Reference to the ``ui.sub_pages`` element currently executing this page.'''
    path_parameters: dict[str, str]
    '''Path parameters extracted from the route pattern.'''
    query_parameters: QueryParams
    '''Query parameters from the request URL.'''
    data: dict[str, Any]
    '''Arbitrary data passed to the ``ui.sub_pages`` element.'''
    remaining_path: str = ''
    '''Remaining path after the matched route (useful for wildcard routing).'''

    @classmethod
    def build_kwargs(cls, match: RouteMatch, frame: SubPages, data: dict[str, Any]) -> dict[str, Any]:
        """Build keyword arguments for the route builder function.

        :param route_match: matched route containing path info and parameters
        :param frame: ``ui.sub_pages`` instance executing this page
        :param data: arbitrary data passed to the ``ui.sub_pages`` element
        :return: keyword arguments for the builder function
        """
        parameters = inspect.signature(match.builder).parameters
        kwargs = {}

        for name, param in parameters.items():
            if param.annotation is cls:
                kwargs[name] = cls._from_route_match(match, frame, data)
            elif name in data:
                kwargs[name] = data[name]
            elif match.parameters and name in match.parameters:
                kwargs[name] = cls._convert_parameter(match.parameters[name], param.annotation)
            elif name in match.query_params:
                kwargs[name] = cls._convert_parameter(match.query_params[name], param.annotation)

        return kwargs

    @classmethod
    def _from_route_match(cls, route_match: RouteMatch, frame: SubPages, data: dict[str, Any]) -> PageArguments:
        """Create a PageArguments instance from route match data.

        :param route_match: matched route containing path info and parameters
        :param frame: SubPages instance executing this page
        :param data: arbitrary data passed to the sub_pages element
        :return: new PageArguments instance
        """
        return cls(
            path=route_match.path,
            frame=frame,
            path_parameters=route_match.parameters or {},
            query_parameters=route_match.query_params,
            data=data,
            remaining_path=route_match.remaining_path,
        )

    @staticmethod
    def _convert_parameter(value: str, param_type: type) -> Any:
        """Convert a string parameter to the specified type (``str``, ``int``, or ``float``).

        :param value: string value to convert
        :param param_type: target type for conversion
        :return: converted value
        """
        param_type = PageArguments._unwrap_optional(param_type)
        if param_type is str or param_type is inspect.Parameter.empty:
            return value
        elif param_type is int:
            return int(value)
        elif param_type is float:
            return float(value)
        return value

    @staticmethod
    def _unwrap_optional(param_type: type) -> type:
        """Extract the base type from T|None -> T, or return the type as-is."""
        if get_origin(param_type) is Union and types.NoneType in get_args(param_type):
            return next(arg for arg in get_args(param_type) if arg is not types.NoneType)
        return param_type



from __future__ import annotations

import asyncio
import time
from pathlib import Path
from typing import TYPE_CHECKING, Any

import httpx

from .. import background_tasks
from ..functions.download import Download

if TYPE_CHECKING:
    from .user import User


class UserDownload(Download):

    def __init__(self, user: User) -> None:
        self.http_responses: list[httpx.Response] = []
        self.user = user

    def __call__(self, src: str | Path | bytes, filename: str | None = None, media_type: str = '') -> Any:
        background_tasks.create(self._get(src),
                                name=f'download {str(src[:10]) + "..." if isinstance(src, bytes) else src}')

    def file(self, path: str | Path, filename: str | None = None, media_type: str = '') -> None:
        self(path)

    def from_url(self, url: str, filename: str | None = None, media_type: str = '') -> None:
        self(url)

    def content(self, content: bytes | str, filename: str | None = None, media_type: str = '') -> None:
        self(content)

    async def _get(self,  src: str | Path | bytes) -> None:
        if isinstance(src, bytes):
            await asyncio.sleep(0)
            response = httpx.Response(httpx.codes.OK, content=src)
        else:
            response = await self.user.http_client.get(str(src))
        self.http_responses.append(response)

    async def next(self, *, timeout: float = 1.0) -> httpx.Response:
        """Wait for a new download to happen.

        :param timeout: the maximum time to wait (default: 1.0)
        :returns: the HTTP response
        """
        assert self.user.client
        downloads = len(self.http_responses)
        deadline = time.time() + timeout
        while len(self.http_responses) < downloads + 1:
            await asyncio.sleep(0.1)
            if time.time() > deadline:
                raise TimeoutError('Download did not happen')
        return self.http_responses[-1]



from .. import helpers, optional_features
from ..element import Element
from .markdown import Markdown

try:
    from nicegui_highcharts import highchart
    optional_features.register('highcharts')
    __all__ = ['highchart']
except ImportError:
    class highchart(Element):  # type: ignore
        def __init__(self, *args, **kwargs) -> None:  # pylint: disable=unused-argument
            """Highcharts chart

            An element to create a chart using `Highcharts <https://www.highcharts.com/>`_.
            Updates can be pushed to the chart by changing the `options` property.

            Due to Highcharts' restrictive license, this element is not part of the standard NiceGUI package.
            It is maintained in a `separate repository <https://github.com/zauberzeug/nicegui-highcharts/>`_
            and can be installed with `pip install nicegui[highcharts]`.
            """
            super().__init__()
            Markdown('Highcharts is not installed. Please run `pip install nicegui[highcharts]`.')
            helpers.warn_once('Highcharts is not installed. Please run "pip install nicegui[highcharts]".')



from ..defaults import DEFAULT_PROP, resolve_defaults
from ..events import Handler, ValueChangeEventArguments
from .mixins.value_element import ValueElement


class Fullscreen(ValueElement, component='fullscreen.js'):
    LOOPBACK = None

    @resolve_defaults
    def __init__(self, *,
                 require_escape_hold: bool = DEFAULT_PROP | False,
                 on_value_change: Handler[ValueChangeEventArguments] | None = None) -> None:
        """Fullscreen control element

        This element is based on Quasar's `AppFullscreen <https://quasar.dev/quasar-plugins/app-fullscreen>`_ plugin
        and provides a way to enter, exit and toggle the fullscreen mode.

        Important notes:

        * Due to security reasons, the fullscreen mode can only be entered from a previous user interaction such as a button click.
        * The long-press escape requirement only works in some browsers like Google Chrome or Microsoft Edge.

        *Added in version 2.11.0*

        :param require_escape_hold: whether the user needs to long-press the escape key to exit fullscreen mode
        :param on_value_change: callback which is invoked when the fullscreen state changes
        """
        super().__init__(value=False, on_value_change=on_value_change)
        self._props['require-escape-hold'] = require_escape_hold

        self._props.add_rename('requireEscapeHold', 'require-escape-hold')  # DEPRECATED: remove in NiceGUI 4.0

    @property
    def require_escape_hold(self) -> bool:
        """Whether the user needs to long-press of the escape key to exit fullscreen mode.

        This feature is only supported in some browsers like Google Chrome or Microsoft Edge.
        In unsupported browsers, this setting has no effect.
        """
        return self._props['require-escape-hold']

    @require_escape_hold.setter
    def require_escape_hold(self, value: bool) -> None:
        self._props['require-escape-hold'] = value

    def enter(self) -> None:
        """Enter fullscreen mode."""
        self.value = True

    def exit(self) -> None:
        """Exit fullscreen mode."""
        self.value = False

    def toggle(self) -> None:
        """Toggle fullscreen mode."""
        self.value = not self.value

    def _handle_value_change(self, value: bool) -> None:
        super()._handle_value_change(value)
        self.run_method('enter' if value else 'exit')



import asyncio
import weakref
from typing import Any

from ..context import context
from ..defaults import DEFAULT_PROPS, resolve_defaults
from ..element import Element
from .mixins.value_element import ValueElement


class Dialog(ValueElement, component='dialog.js'):

    @resolve_defaults
    def __init__(self, *, value: bool = DEFAULT_PROPS['model-value'] | False) -> None:
        """Dialog

        Creates a dialog based on Quasar's `QDialog <https://quasar.dev/vue-components/dialog>`_ component.
        By default it is dismissible by clicking or pressing ESC.
        To make it persistent, set `.props('persistent')` on the dialog element.

        NOTE: The dialog is an element.
        That means it is not removed when closed, but only hidden.
        You should either create it only once and then reuse it, or remove it with `.clear()` after dismissal.

        :param value: whether the dialog should be opened on creation (default: `False`)
        """
        with context.client.layout:
            super().__init__(value=value, on_value_change=None)

        # create a canary element in the current context to trigger the deletion of the dialog when its parent is deleted
        canary = Element()
        canary.visible = False
        weakref.finalize(
            canary, lambda: self.delete() if not self.is_deleted and self._parent_slot and self._parent_slot() else None
        )

        self._result: Any = None
        self._submitted: asyncio.Event | None = None

    @property
    def submitted(self) -> asyncio.Event:
        """An event that is set when the dialog is submitted."""
        if self._submitted is None:
            self._submitted = asyncio.Event()
        return self._submitted

    def open(self) -> None:
        """Open the dialog."""
        self.value = True

    def close(self) -> None:
        """Close the dialog."""
        self.value = False

    def __await__(self):
        self._result = None
        self.submitted.clear()
        self.open()
        yield from self.submitted.wait().__await__()  # pylint: disable=no-member
        result = self._result
        self.close()
        return result

    def submit(self, result: Any) -> None:
        """Submit the dialog with the given result."""
        self._result = result
        self.submitted.set()

    def _handle_value_change(self, value: Any) -> None:
        super()._handle_value_change(value)
        if not self.value:
            self._result = None
            self.submitted.set()



import ast
import re
import weakref
from collections.abc import Iterator
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from . import helpers
from .observables import ObservableDict

if TYPE_CHECKING:
    from .element import Element

PROPS_PATTERN = re.compile(r'''
# Match a key or key-value pair optionally followed by whitespace or end of string
# The value can be unquoted, or enclosed in quotes, or square or curly brackets
(?P<key>[:\w\-]+)           # Capture group 1: Key
(                           # Optional non-capturing group for value
    =                       # An equals sign
    (                       # followed by one of ...
        (?P<double>         # a value enclosed in double quotes
            "                   # Match double quote
            [^"\\]*             # Match any character except quotes or backslashes zero or more times
            (\\.[^"\\]*)*       # Match any escaped character followed by any character except quotes or backslashes zero or more times
            "                   # Match the closing quote
        )
        |                   # or
        (?P<single>         # a value enclosed in single quotes
            '                   # Match a single quote
            [^'\\]*             # Match any character except quotes or backslashes zero or more times
            (\\.[^'\\]*)*       # Match any escaped character followed by any character except quotes or backslashes zero or more times
            '                   # Match the closing quote
        )
        |                   # or
        (?P<square>         # a value enclosed in square brackets
            \[                  # Match the opening [
            [^\]\\]*            # Match any character except a ] or backslashes zero or more times
            (\\.[^\]\\]*)*      # Match any escaped character followed by any character except ] or backslashes zero or more times
            \]                  # Match the closing ]
        )
        |                   # or
        (?P<curly>          # a value enclosed in curly braces
            \{                  # Match the opening {
            [^\}\\]*            # Match any character except } or backslashes zero or more times
            (\\.[^\}\\]*)*      # Match any escaped character followed by any character except ] or backslashes zero or more times
            \}                  # Match the closing }
        )
        |                   # or, as a final alternative ....
        (?P<unquoted>[\w\-.,%:\/=?&;+#@~$]+)  # a value without quotes
    )
)?                          # End of optional non-capturing group for value
($|\s)                      # Match end of string or whitespace
''', re.VERBOSE)

T = TypeVar('T', bound='Element')


class Props(ObservableDict, Generic[T]):

    def __init__(self, *args, element: T, **kwargs) -> None:
        super().__init__(*args, on_change=self._update, **kwargs)
        self._element = weakref.ref(element)
        self._warnings: dict[str, str] = {}
        self._renames: dict[str, str] = {}
        self._suspend_count = 0

    def set_optional(self, key: str, value: Any) -> None:
        """Set a prop value or remove it if None is provided."""
        if value is None:
            self.pop(key, None)
        else:
            self[key] = value

    def set_bool(self, key: str, value: Any) -> None:
        """Set a boolean prop value or remove it if falsy value is provided."""
        if value:
            self[key] = value
        else:
            self.pop(key, None)

    @contextmanager
    def suspend_updates(self) -> Iterator[None]:
        """Suspend updates."""
        self._suspend_count += 1
        try:
            yield
        finally:
            self._suspend_count -= 1

    @property
    def element(self) -> T:
        """The element this props object belongs to."""
        element = self._element()
        if element is None:
            raise RuntimeError('The element this props object belongs to has been deleted.')
        return element

    def _update(self) -> None:
        if self._suspend_count > 0:
            return

        for name, message in self._warnings.items():
            self._check_warning(name, message)

        for name, rename in self._renames.items():
            self._check_rename(name, rename)

        element = self._element()
        if element is not None:
            element.update()

    def add_warning(self, prop: str, message: str) -> None:
        """Add a warning message for a prop."""
        self._warnings[prop] = message
        self._check_warning(prop, message)

    def add_rename(self, prop: str, replacement: str) -> None:
        """Add a rename warning for a prop."""
        self._renames[prop] = replacement
        self._check_rename(prop, replacement)

    def _check_warning(self, name: str, message: str) -> None:
        if name in self:
            with self.suspend_updates():
                del self[name]
            helpers.warn_once(message)

    def _check_rename(self, name: str, rename: str) -> None:
        if name in self:
            with self.suspend_updates():
                self[rename] = self[name]
            helpers.warn_once(f'The prop "{name}" is deprecated. Use "{rename}" instead.')

    def __call__(self,
                 add: str | None = None, *,
                 remove: str | None = None) -> T:
        """Add or remove props.

        This allows modifying the look of the element or its layout using `Quasar <https://quasar.dev/>`_ props.
        Since props are simply applied as HTML attributes, they can be used with any HTML element.

        Boolean properties are assumed ``True`` if no value is specified.

        :param add: whitespace-delimited list of either boolean values or key=value pair to add
        :param remove: whitespace-delimited list of property keys to remove
        """
        element = self.element
        for key in self.parse(remove):
            if key in self:
                del self[key]
        for key, value in self.parse(add).items():
            if self.get(key) != value:
                self[key] = value
        return element

    @staticmethod
    def parse(text: str | None) -> dict[str, Any]:
        """Parse a string of props into a dictionary."""
        if not text:
            return {}
        props: dict[str, Any] = {}
        for match in PROPS_PATTERN.finditer(text):
            match_groups = match.groupdict()
            key = match_groups['key']

            for group in ['single', 'double', 'square', 'curly']:
                if match_groups[group] is not None:
                    props[key] = ast.literal_eval(match_groups[group])
                    break
            else:
                if match_groups['unquoted'] is not None:
                    props[key] = match_groups['unquoted']
                else:
                    props[key] = True

        return props



from ..element import Element


class Grid(Element, default_classes='nicegui-grid'):

    def __init__(self,
                 *,
                 rows: int | str | None = None,
                 columns: int | str | None = None,
                 ) -> None:
        """Grid Element

        Provides a container which arranges its child in a grid.

        :param rows: number of rows in the grid or a string with the grid-template-rows CSS property (e.g. 'auto 1fr')
        :param columns: number of columns in the grid or a string with the grid-template-columns CSS property (e.g. 'auto 1fr')
        """
        super().__init__('div')

        if isinstance(rows, int):
            self._style['grid-template-rows'] = f'repeat({rows}, minmax(0, 1fr))'
        elif isinstance(rows, str):
            self._style['grid-template-rows'] = rows

        if isinstance(columns, int):
            self._style['grid-template-columns'] = f'repeat({columns}, minmax(0, 1fr))'
        elif isinstance(columns, str):
            self._style['grid-template-columns'] = columns



from typing import Literal

from typing_extensions import Self

from ..element import Element
from ..events import GenericEventArguments, Handler, ScrollEventArguments, handle_event


class ScrollArea(Element, default_classes='nicegui-scroll-area'):

    def __init__(self, *, on_scroll: Handler[ScrollEventArguments] | None = None) -> None:
        """Scroll Area

        A way of customizing the scrollbars by encapsulating your content.
        This element exposes the Quasar `ScrollArea <https://quasar.dev/vue-components/scroll-area/>`_ component.

        :param on_scroll: function to be called when the scroll position changes
        """
        super().__init__('q-scroll-area')

        if on_scroll:
            self.on_scroll(on_scroll)

    def on_scroll(self, callback: Handler[ScrollEventArguments]) -> Self:
        """Add a callback to be invoked when the scroll position changes."""
        self.on('scroll', lambda e: self._handle_scroll(callback, e), args=[
            'verticalPosition',
            'verticalPercentage',
            'verticalSize',
            'verticalContainerSize',
            'horizontalPosition',
            'horizontalPercentage',
            'horizontalSize',
            'horizontalContainerSize',
        ])
        return self

    def _handle_scroll(self, handler: Handler[ScrollEventArguments] | None, e: GenericEventArguments) -> None:
        handle_event(handler, ScrollEventArguments(
            sender=self,
            client=self.client,
            vertical_position=e.args['verticalPosition'],
            vertical_percentage=e.args['verticalPercentage'],
            vertical_size=e.args['verticalSize'],
            vertical_container_size=e.args['verticalContainerSize'],
            horizontal_position=e.args['horizontalPosition'],
            horizontal_percentage=e.args['horizontalPercentage'],
            horizontal_size=e.args['horizontalSize'],
            horizontal_container_size=e.args['horizontalContainerSize'],
        ))

    def scroll_to(self, *,
                  pixels: float | None = None,
                  percent: float | None = None,
                  axis: Literal['vertical', 'horizontal'] = 'vertical',
                  duration: float = 0.0,
                  ) -> None:
        """Set the scroll area position in percentage (float) or pixel number (int).

        You can add a delay to the actual scroll action with the `duration_ms` parameter.

        :param pixels: scroll position offset from top in pixels
        :param percent: scroll position offset from top in percentage of the total scrolling size
        :param axis: scroll axis
        :param duration: animation duration (in seconds, default: 0.0 means no animation)
        """
        if pixels is not None and percent is not None:
            raise ValueError('You can only specify one of pixels or percent')
        if pixels is not None:
            self.run_method('setScrollPosition', axis, pixels, 1000 * duration)
        elif percent is not None:
            self.run_method('setScrollPercentage', axis, percent, 1000 * duration)
        else:
            raise ValueError('You must specify one of pixels or percent')



from pathlib import Path

from .. import core, helpers
from ..context import context
from ..logging import log


class Download:
    """Download functions

    These functions allow you to download files, URLs or raw data.

    *Added in version 2.14.0*
    """

    def __call__(self, src: str | Path | bytes, filename: str | None = None, media_type: str = '') -> None:
        """Download

        Function to trigger the download of a file, URL or bytes.

        :param src: relative target URL, local path of a file or raw data which should be downloaded
        :param filename: name of the file to download (default: name of the file on the server)
        :param media_type: media type of the file to download (default: "")
        """
        if isinstance(src, bytes):
            self.content(src, filename, media_type)
        elif helpers.is_file(src):
            self.file(src, filename, media_type)
        else:
            src = str(src)
            self.from_url(src, filename, media_type)

    def file(self, path: str | Path, filename: str | None = None, media_type: str = '') -> None:
        """Download file from local path

        Function to trigger the download of a file.

        *Added in version 2.14.0*

        :param path: local path of the file
        :param filename: name of the file to download (default: name of the file on the server)
        :param media_type: media type of the file to download (default: "")
        """
        src = core.app.add_static_file(local_file=path, single_use=True)
        context.client.download(src, filename, media_type)

    def from_url(self, url: str, filename: str | None = None, media_type: str = '') -> None:
        """Download from a relative URL

        Function to trigger the download from a relative URL.

        Note:
        This function is intended to be used with relative URLs only.
        For absolute URLs, the browser ignores the download instruction and tries to view the file in a new tab
        if possible, such as images, PDFs, etc.
        Therefore, the download may only work for some file types such as .zip, .db, etc.
        Furthermore, the browser ignores filename and media_type parameters,
        respecting the origin server's headers instead.
        Either replace the absolute URL with a relative one, or use ``ui.navigate.to(url, new_tab=True)`` instead.

        *Added in version 2.14.0*

        *Updated in version 2.19.0: Added warning for cross-origin downloads*

        :param url: URL
        :param filename: name of the file to download (default: name of the file on the server)
        :param media_type: media type of the file to download (default: "")
        """
        is_relative = url.startswith('/') or url.startswith('./') or url.startswith('../')
        if not is_relative:
            log.warning('Using `ui.download.from_url` with absolute URLs is not recommended.\n'
                        'Please refer to the documentation for more details.')
        context.client.download(url, filename, media_type)

    def content(self, content: bytes | str, filename: str | None = None, media_type: str = '') -> None:
        """Download raw bytes or string content

        Function to trigger the download of raw data.

        *Added in version 2.14.0*

        :param content: raw bytes or string
        :param filename: name of the file to download (default: name of the file on the server)
        :param media_type: media type of the file to download (default: "")
        """
        if isinstance(content, str):
            content = content.encode('utf-8')
        context.client.download(content, filename, media_type)


download = Download()



import os
import shutil
from collections.abc import Generator
from pathlib import Path

import pytest
from selenium import webdriver
from selenium.webdriver.chrome.service import Service

from .general_fixtures import (  # noqa: F401  # pylint: disable=unused-import
    nicegui_reset_globals,
    pytest_addoption,
    pytest_configure,
)
from .screen import Screen

# pylint: disable=redefined-outer-name

DOWNLOAD_DIR = Path(__file__).parent / 'download'


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):  # pylint: disable=unused-argument
    """Store test outcome in the node for fixture access."""
    outcome = yield
    rep = outcome.get_result()
    setattr(item, f'rep_{rep.when}', rep)


@pytest.fixture(scope='session')
def nicegui_chrome_options() -> webdriver.ChromeOptions:
    """Configure the Chrome options for the NiceGUI tests."""
    chrome_options = webdriver.ChromeOptions()
    chrome_options.add_argument('disable-dev-shm-usage')
    chrome_options.add_argument('disable-search-engine-choice-screen')
    chrome_options.add_argument('no-sandbox')
    chrome_options.add_argument('headless')
    if 'GITHUB_ACTIONS' in os.environ:
        chrome_options.add_argument('disable-gpu')
        chrome_options.add_argument('enable-unsafe-swiftshader')
    else:
        chrome_options.add_argument('--use-gl=angle')
    chrome_options.add_argument('window-size=600x600')
    chrome_options.add_experimental_option('prefs', {
        'download.default_directory': str(DOWNLOAD_DIR),
        'download.prompt_for_download': False,  # To auto download the file
        'download.directory_upgrade': True,
    })
    chrome_options.set_capability('goog:loggingPrefs', {'browser': 'ALL'})
    if 'CHROME_BINARY_LOCATION' in os.environ:
        chrome_options.binary_location = os.environ['CHROME_BINARY_LOCATION']
    return chrome_options


@pytest.fixture(scope='session')
def nicegui_remove_all_screenshots() -> None:
    """Remove all screenshots from the screenshot directory before the test session."""
    for name in Screen.SCREENSHOT_DIR.glob('*.png'):
        name.unlink()


@pytest.fixture(scope='session')
def nicegui_driver(nicegui_chrome_options: webdriver.ChromeOptions) -> Generator[webdriver.Chrome, None, None]:
    """Create a new Chrome driver instance (reused across tests in the session)."""
    for executable_path in (None, shutil.which('chromedriver'), 'chromedriver'):  # Required for ARM devcontainers
        try:
            s = Service(executable_path=executable_path)
            driver_ = webdriver.Chrome(service=s, options=nicegui_chrome_options)
            break
        except Exception:
            continue
    else:  # no break
        raise RuntimeError('Could not start Chrome WebDriver.')
    driver_.implicitly_wait(Screen.IMPLICIT_WAIT)
    driver_.set_page_load_timeout(4)
    yield driver_
    driver_.quit()


@pytest.fixture
def screen(nicegui_reset_globals,  # noqa: F811, pylint: disable=unused-argument
           nicegui_remove_all_screenshots,  # pylint: disable=unused-argument
           nicegui_driver: webdriver.Chrome,
           request: pytest.FixtureRequest,
           caplog: pytest.LogCaptureFixture,
           ) -> Generator[Screen, None, None]:
    """Create a new SeleniumScreen fixture."""
    _reset_browser_state(nicegui_driver)
    os.environ['NICEGUI_SCREEN_TEST_PORT'] = str(Screen.PORT)
    screen_ = Screen(nicegui_driver, caplog, request)
    try:
        yield screen_

        logs = [record for record in screen_.caplog.get_records('call') if record.levelname == 'ERROR']
        if screen_.is_open:
            test_failed = hasattr(request.node, 'rep_call') and request.node.rep_call.failed
            screen_.shot(request.node.name, failed=test_failed or bool(logs))
        if logs:
            pytest.fail('There were unexpected ERROR logs.', pytrace=False)
        if screen_.is_open and Screen.CATCH_JS_ERRORS:
            for js_error in screen_.selenium.get_log('browser'):
                if str(js_error.get('level', '')).upper() in ('SEVERE', 'ERROR') and \
                        not any(allowed_error in js_error['message'] for allowed_error in screen_.allowed_js_errors):
                    pytest.fail(f'JavaScript console error:\n{js_error}', pytrace=False)
    finally:
        os.environ.pop('NICEGUI_SCREEN_TEST_PORT', None)
        screen_.stop_server()
        if DOWNLOAD_DIR.exists():
            shutil.rmtree(DOWNLOAD_DIR)


def _reset_browser_state(driver: webdriver.Chrome) -> None:
    """Reset browser state between tests when reusing the driver."""
    while len(driver.window_handles) > 1:
        driver.switch_to.window(driver.window_handles[-1])
        driver.close()
    driver.switch_to.window(driver.window_handles[0])
    driver.execute_cdp_cmd('Storage.clearDataForOrigin', {
        'origin': f'http://localhost:{Screen.PORT}',
        'storageTypes': 'cookies,local_storage,session_storage',
    })
    driver.get('about:blank')
    driver.get_log('browser')



from typing_extensions import Self

from ..defaults import DEFAULT_PROPS, resolve_defaults
from ..element import Element
from ..events import ColorPickEventArguments, GenericEventArguments, Handler, handle_event
from .menu import Menu


class ColorPicker(Menu):

    @resolve_defaults
    def __init__(self, *,
                 on_pick: Handler[ColorPickEventArguments] | None = None,
                 value: bool = DEFAULT_PROPS['model-value'] | False,
                 ) -> None:
        """Color Picker

        This element is based on Quasar's `QMenu <https://quasar.dev/vue-components/menu>`_ and
        `QColor <https://quasar.dev/vue-components/color-picker>`_ components.

        :param on_pick: callback to execute when a color is picked
        :param value: whether the menu is already opened (default: `False`)
        """
        super().__init__(value=value)
        self._pick_handlers = [on_pick] if on_pick else []
        with self:
            def handle_change(e: GenericEventArguments):
                for handler in self._pick_handlers:
                    handle_event(handler, ColorPickEventArguments(sender=self, client=self.client, color=e.args))
            self.q_color = Element('q-color').on('change', handle_change)

    def set_color(self, color: str) -> None:
        """Set the color of the picker.

        :param color: the color to set
        """
        self.q_color.props(f'model-value="{color}"')

    def on_pick(self, callback: Handler[ColorPickEventArguments]) -> Self:
        """Add a callback to be invoked when a color is picked."""
        self._pick_handlers.append(callback)
        return self



from collections.abc import Callable
from typing import Any, cast

from typing_extensions import Self

from ...binding import BindableProperty, bind, bind_from, bind_to
from ...element import Element
from ...events import Handler, ValueChangeEventArguments, handle_event


class SelectableElement(Element):
    selected = BindableProperty(
        on_change=lambda sender, selected: cast(Self, sender)._handle_selection_change(selected))  # pylint: disable=protected-access

    def __init__(self, *,
                 selectable: bool,
                 selected: bool,
                 on_selection_change: Handler[ValueChangeEventArguments] | None = None,
                 **kwargs: Any) -> None:
        super().__init__(**kwargs)
        if not selectable:
            return

        self._props['selectable'] = selectable

        self.selected = selected
        self._props['selected'] = selected
        self.set_selected(selected)
        self.on('update:selected', lambda e: self.set_selected(e.args))

        self._selection_change_handlers: list[Handler[ValueChangeEventArguments]] = []
        if on_selection_change:
            self.on_selection_change(on_selection_change)

    def on_selection_change(self, callback: Handler[ValueChangeEventArguments]) -> Self:
        """Add a callback to be invoked when the selection state changes."""
        self._selection_change_handlers.append(callback)
        return self

    def bind_selected_to(self,
                         target_object: Any,
                         target_name: str = 'selected',
                         forward: Callable[[Any], Any] | None = None, *,
                         strict: bool | None = None,
                         ) -> Self:
        """Bind the selection state of this element to the target object's target_name property.

        The binding works one way only, from this element to the target.
        The update happens immediately and whenever a value changes.

        :param target_object: The object to bind to.
        :param target_name: The name of the property to bind to.
        :param forward: A function to apply to the value before applying it to the target (default: identity).
        :param strict: Whether to check (and raise) if the target object has the specified property (default: None,
            performs a check if the object is not a dictionary, *added in version 3.0.0*).
        """
        bind_to(self, 'selected', target_object, target_name, forward, self_strict=False, other_strict=strict)
        return self

    def bind_selected_from(self,
                           target_object: Any,
                           target_name: str = 'selected',
                           backward: Callable[[Any], Any] | None = None, *,
                           strict: bool | None = None,
                           ) -> Self:
        """Bind the selection state of this element from the target object's target_name property.

        The binding works one way only, from the target to this element.
        The update happens immediately and whenever a value changes.

        :param target_object: The object to bind from.
        :param target_name: The name of the property to bind from.
        :param backward: A function to apply to the value before applying it to this element (default: identity).
        :param strict: Whether to check (and raise) if the target object has the specified property (default: None,
            performs a check if the object is not a dictionary, *added in version 3.0.0*).
        """
        bind_from(self, 'selected', target_object, target_name, backward, self_strict=False, other_strict=strict)
        return self

    def bind_selected(self,
                      target_object: Any,
                      target_name: str = 'selected', *,
                      forward: Callable[[Any], Any] | None = None,
                      backward: Callable[[Any], Any] | None = None,
                      strict: bool | None = None,
                      ) -> Self:
        """Bind the selection state of this element to the target object's target_name property.

        The binding works both ways, from this element to the target and from the target to this element.
        The update happens immediately and whenever a value changes.
        The backward binding takes precedence for the initial synchronization.

        :param target_object: The object to bind to.
        :param target_name: The name of the property to bind to.
        :param forward: A function to apply to the value before applying it to the target (default: identity).
        :param backward: A function to apply to the value before applying it to this element (default: identity).
        :param strict: Whether to check (and raise) if the target object has the specified property (default: None,
            performs a check if the object is not a dictionary, *added in version 3.0.0*).
        """
        bind(self, 'selected', target_object, target_name,
             forward=forward, backward=backward,
             self_strict=False, other_strict=strict)
        return self

    def set_selected(self, selected: bool) -> None:
        """Set the selection state of this element.

        :param selected: The new selection state.
        """
        self.selected = selected

    def _handle_selection_change(self, selected: bool) -> None:
        """Called when the selection state of this element changes.

        :param selected: The new selection state.
        """
        previous_value = self._props.get('selected')
        self._props['selected'] = selected
        args = ValueChangeEventArguments(sender=self, client=self.client, value=selected, previous_value=previous_value)
        for handler in self._selection_change_handlers:
            handle_event(handler, args)



from __future__ import annotations

from typing import Any, cast

from ..context import context
from ..defaults import DEFAULT_PROP, DEFAULT_PROPS, resolve_defaults
from ..element import Element
from ..events import Handler, ValueChangeEventArguments
from .mixins.disableable_element import DisableableElement
from .mixins.icon_element import IconElement
from .mixins.value_element import ValueElement


class Stepper(ValueElement, default_classes='nicegui-stepper'):

    @resolve_defaults
    def __init__(self, *,
                 value: str | Step | None = DEFAULT_PROPS['model-value'] | None,
                 on_value_change: Handler[ValueChangeEventArguments] | None = None,
                 keep_alive: bool = DEFAULT_PROP | True,
                 ) -> None:
        """Stepper

        This element represents `Quasar's QStepper <https://quasar.dev/vue-components/stepper#qstepper-api>`_ component.
        It contains individual steps.

        To avoid issues with dynamic elements when switching steps,
        this element uses Vue's `keep-alive <https://vuejs.org/guide/built-ins/keep-alive.html>`_ component.
        If client-side performance is an issue, you can disable this feature.

        :param value: `ui.step` or name of the step to be initially selected (default: `None` meaning the first step)
        :param on_value_change: callback to be executed when the selected step changes
        :param keep_alive: whether to use Vue's keep-alive component on the content (default: `True`)
        """
        super().__init__(tag='q-stepper', value=value, on_value_change=on_value_change)
        self._props.set_bool('keep-alive', keep_alive)

    def _value_to_model_value(self, value: Any) -> Any:
        return value.props['name'] if isinstance(value, Step) else value

    def _handle_value_change(self, value: Any) -> None:
        super()._handle_value_change(value)
        names = [step.props['name'] for step in self]
        for i, step in enumerate(self):
            done = i < names.index(value) if value in names else False
            step.props(f':done={done}')

    def next(self) -> None:
        """Show the next step."""
        self.run_method('next')

    def previous(self) -> None:
        """Show the previous step."""
        self.run_method('previous')


class Step(IconElement, DisableableElement, default_classes='nicegui-step'):

    @resolve_defaults
    def __init__(self, name: str, title: str | None = None, icon: str | None = DEFAULT_PROP | None) -> None:
        """Step

        This element represents `Quasar's QStep <https://quasar.dev/vue-components/stepper#qstep-api>`_ component.
        It is a child of a `ui.stepper` element.

        :param name: name of the step (will be the value of the `ui.stepper` element)
        :param title: title of the step (default: `None`, meaning the same as `name`)
        :param icon: icon of the step (default: `None`)
        """
        super().__init__(tag='q-step', icon=icon)
        self._props['name'] = name
        self._props['title'] = title if title is not None else name
        self.stepper = cast(ValueElement, context.slot.parent)
        if self.stepper.value is None:
            self.stepper.value = name


class StepperNavigation(Element):

    def __init__(self, *, wrap: bool = True) -> None:
        """Stepper Navigation

        This element represents `Quasar's QStepperNavigation https://quasar.dev/vue-components/stepper#qsteppernavigation-api>`_ component.

        :param wrap: whether to wrap the content (default: `True`)
        """
        super().__init__('q-stepper-navigation')

        if wrap:
            self._classes.append('wrap')
